@@ -100,6 +100,45 @@ include the pathname of the file and the position of the definition."
 FBOUNDP."
   (and (sb-int:valid-function-name-p name) t))
 
+;;;; Utilities for code
+
+(declaim (inline map-code-constants))
+(defun map-code-constants (code fn)
+  "Call FN for each constant in CODE's constant pool."
+  (check-type code sb-kernel:code-component)
+  (loop for i from sb-vm:code-constants-offset below
+       (sb-kernel:get-header-data code)
+     do (funcall fn (sb-kernel:code-header-ref code i))))
+
+(declaim (inline map-allocated-code-components))
+(defun map-allocated-code-components (spaces fn)
+  "Call FN for each allocated code component in one of SPACES.  FN
+receives the object and its size as arguments.  SPACES should be a
+list of the symbols :dynamic, :static, or :read-only."
+  (dolist (space spaces)
+    (sb-vm::map-allocated-objects
+     (lambda (obj header size)
+       (when (= sb-vm:code-header-widetag header)
+         (funcall fn obj size)))
+     space)))
+
+(declaim (inline map-caller-code-components))
+(defun map-caller-code-components (function spaces fn)
+  "Call FN for each code component with a fdefn for FUNCTION in its
+constant pool."
+  (let ((function (coerce function 'function)))
+    (map-allocated-code-components
+     spaces
+     (lambda (obj size)
+       (declare (ignore size))
+       (map-code-constants
+        obj
+        (lambda (constant)
+          (when (and (sb-kernel:fdefn-p constant)
+                     (eq (sb-kernel:fdefn-fun constant)
+                         function))
+            (funcall fn obj))))))))
+
 ;;;; Finding definitions
 
 (defstruct definition-source
@@ -123,30 +162,6 @@ FBOUNDP."
   ;; is.
   (description nil :type list))
 
-(defun vop-sources-from-fun-templates (name)
-  (let ((fun-info (sb-int:info :function :info name)))
-    (when fun-info
-      (loop for vop in (sb-c::fun-info-templates fun-info)
-            for source = (find-definition-source
-                          (sb-c::vop-info-generator-function vop))
-            do (setf (definition-source-description source)
-                     (list (sb-c::template-name vop)
-                           (sb-c::template-note vop)))
-            collect source))))
-
-(defun find-vop-source (name)
-  (let* ((templates (vop-sources-from-fun-templates name))
-         (vop (gethash name sb-c::*backend-template-names*))
-         (source (and vop
-                      (find-definition-source
-                       (sb-c::vop-info-generator-function vop)))))
-    (when source
-      (setf (definition-source-description source)
-            (list name)))
-    (if source
-        (cons source templates)
-        templates)))
-
 (defun find-definition-sources-by-name (name type)
   "Returns a list of DEFINITION-SOURCEs for the objects of type TYPE
 defined with name NAME. NAME may be a symbol or a extended function
@@ -167,13 +182,16 @@ name. Type can currently be one of the following:
    :STRUCTURE
    :SYMBOL-MACRO
    :TYPE
+   :ALIEN-TYPE
    :VARIABLE
+   :DECLARATION
 
    (Internal)
    :OPTIMIZER
    :SOURCE-TRANSFORM
    :TRANSFORM
    :VOP
+   :IR1-CONVERT
 
 If an unsupported TYPE is requested, the function will return NIL.
 "
@@ -196,7 +214,7 @@ If an unsupported TYPE is requested, the function will return NIL.
        ((:variable)
         (when (and (symbolp name)
                    (member (sb-int:info :variable :kind name)
-                           '(:global :special)))
+                           '(:global :special :alien)))
           (translate-source-location (sb-int:info :source-location type name))))
        ((:constant)
         (when (and (symbolp name)
@@ -213,11 +231,16 @@ If an unsupported TYPE is requested, the function will return NIL.
        ((:compiler-macro)
         (when (compiler-macro-function name)
           (find-definition-source (compiler-macro-function name))))
+       (:ir1-convert
+        (let ((converter (sb-int:info :function :ir1-convert name)))
+          (and converter
+           (find-definition-source converter))))
        ((:function :generic-function)
         (when (and (fboundp name)
-                   (or (not (symbolp name))
-                       (not (macro-function name))
-                       (special-operator-p name)))
+                   (or (consp name)
+                       (and
+                        (not (macro-function name))
+                        (not (special-operator-p name)))))
           (let ((fun (real-fdefinition name)))
             (when (eq (not (typep fun 'generic-function))
                       (not (eq type :generic-function)))
@@ -312,22 +335,32 @@ If an unsupported TYPE is requested, the function will return NIL.
                       (setf (definition-source-description source)
                             (list name))
                       source))))))
-       ((:vop)
-        (when (symbolp name)
-          (find-vop-source name)))
+       ((:vop :alien-type)
+        (let ((loc (sb-int:info :source-location type name)))
+          (and loc
+               (translate-source-location loc))))
        ((:source-transform)
-        (let ((transform-fun (sb-int:info :function :source-transform name))
-              (accessor (sb-int:info :function :structure-accessor
-                                     (if (typep name '(cons (eql setf)
-                                                       (cons symbol null)))
-                                         (second name)
-                                         name))))
+        (let* ((transform-fun
+                (or (sb-int:info :function :source-transform name)
+                    (and (typep name '(cons (eql setf) (cons symbol null)))
+                         (sb-int:info :function :source-transform
+                                      (second name)))))
+               ;; A cons for the :source-transform is essentially the same
+               ;; info that was formerly in :structure-accessor.
+               (accessor (and (consp transform-fun) (cdr transform-fun))))
           ;; Structure accessors have source transforms, but the
           ;; returned locations will neither show the actual place
           ;; where it's defined, nor is really interesting.
           (when (and transform-fun
                      (not accessor))
             (find-definition-source transform-fun))))
+       (:declaration
+        (let ((locations (sb-int:info :source-location :declaration name)))
+          (loop for (kind loc) on locations by #'cddr
+                when loc
+                collect (let ((loc (translate-source-location loc)))
+                          (setf (definition-source-description loc) (list kind))
+                          loc))))
        (t
         nil)))))
 
@@ -431,7 +464,9 @@ If an unsupported TYPE is requested, the function will return NIL.
 Works for special-operators, macros, simple functions, interpreted functions,
 and generic functions. Signals an error if FUNCTION is not a valid extended
 function designator."
-  (cond ((valid-function-name-p function)
+  (cond ((and (symbolp function) (special-operator-p function))
+         (function-lambda-list (sb-int:info :function :ir1-convert function)))
+        ((valid-function-name-p function)
          (function-lambda-list (or (and (symbolp function)
                                         (macro-function function))
                                    (fdefinition function))))
@@ -439,7 +474,7 @@ function designator."
          (sb-pcl::generic-function-pretty-arglist function))
         #+sb-eval
         ((typep function 'sb-eval:interpreted-function)
-         (sb-eval:interpreted-function-lambda-list function))
+         (sb-eval:interpreted-function-debug-lambda-list function))
         (t
          (sb-kernel:%simple-fun-arglist (sb-kernel:%fun-fun function)))))
 
@@ -448,18 +483,14 @@ function designator."
 value, and a flag whether the arglist could be found as second
 value."
   (check-type typespec-operator symbol)
-  (case (sb-int:info :type :kind typespec-operator)
-    (:defined
-     (sb-int:info :type :lambda-list typespec-operator))
-    (:primitive
-     (let ((translator-fun (sb-int:info :type :translator typespec-operator)))
-       (if translator-fun
-           (values (sb-kernel:%fun-lambda-list translator-fun) t)
-           ;; Some builtin types (e.g. STRING) do not have a
-           ;; translator, but they were actually defined via DEFTYPE
-           ;; in src/code/deftypes-for-target.lisp.
-           (sb-int:info :type :lambda-list typespec-operator))))
-    (t (values nil nil))))
+  ;; Don't return a lambda-list for combinators AND,OR,NOT.
+  (let ((f (and (sb-int:info :type :kind typespec-operator)
+                ;; globaldb prevents storing both of these
+                (or (sb-int:info :type :expander typespec-operator)
+                    (sb-int:info :type :translator typespec-operator)))))
+    (if (functionp f)
+        (values (sb-kernel:%fun-lambda-list f) t)
+        (values nil nil))))
 
 (defun function-type (function-designator)
   "Returns the ftype of FUNCTION-DESIGNATOR, or NIL."
@@ -490,7 +521,7 @@ value."
 
 ;;;; find callers/callees, liberated from Helmut Eller's code in SLIME
 
-;;; This interface is trmendously experimental.
+;;; This interface is tremendously experimental.
 
 ;;; For the moment I'm taking the view that FDEFN is an internal
 ;;; object (one out of one CMUCL developer surveyed didn't know what
@@ -504,6 +535,7 @@ value."
 
 (defun find-function-callees (function)
   "Return functions called by FUNCTION."
+  (declare (sb-kernel:simple-fun function))
   (let ((callees '()))
     (map-code-constants
      (sb-kernel:fun-code-header function)
@@ -513,7 +545,6 @@ value."
                callees))))
     callees))
 
-
 (defun find-function-callers (function &optional (spaces '(:read-only :static
                                                            :dynamic)))
   "Return functions which call FUNCTION, by searching SPACES for code objects"
@@ -531,43 +562,6 @@ value."
                       do (pushnew e referrers)))))))
     referrers))
 
-(declaim (inline map-code-constants))
-(defun map-code-constants (code fn)
-  "Call FN for each constant in CODE's constant pool."
-  (check-type code sb-kernel:code-component)
-  (loop for i from sb-vm:code-constants-offset below
-        (sb-kernel:get-header-data code)
-        do (funcall fn (sb-kernel:code-header-ref code i))))
-
-(declaim (inline map-allocated-code-components))
-(defun map-allocated-code-components (spaces fn)
-  "Call FN for each allocated code component in one of SPACES.  FN
-receives the object and its size as arguments.  SPACES should be a
-list of the symbols :dynamic, :static, or :read-only."
-  (dolist (space spaces)
-    (sb-vm::map-allocated-objects
-     (lambda (obj header size)
-       (when (= sb-vm:code-header-widetag header)
-         (funcall fn obj size)))
-     space)))
-
-(declaim (inline map-caller-code-components))
-(defun map-caller-code-components (function spaces fn)
-  "Call FN for each code component with a fdefn for FUNCTION in its
-constant pool."
-  (let ((function (coerce function 'function)))
-    (map-allocated-code-components
-     spaces
-     (lambda (obj size)
-       (declare (ignore size))
-       (map-code-constants
-        obj
-        (lambda (constant)
-          (when (and (sb-kernel:fdefn-p constant)
-                     (eq (sb-kernel:fdefn-fun constant)
-                         function))
-            (funcall fn obj))))))))
-
 ;;; XREF facility
 
 (defun get-simple-fun (functoid)
@@ -906,13 +900,9 @@ Experimental: interface subject to change."
          (call (realpart object))
          (call (realpart object)))
         (sb-vm::instance
-         (let* ((len (sb-kernel:%instance-length object))
-                (nuntagged (if (typep object 'structure-object)
-                               (sb-kernel:layout-n-untagged-slots
-                                (sb-kernel:%instance-layout object))
-                               0)))
-           (dotimes (i (- len nuntagged))
-             (call (sb-kernel:%instance-ref object i))))
+         (call (sb-kernel:%instance-layout object))
+         (sb-kernel:do-instance-tagged-slot (i object)
+           (call (sb-kernel:%instance-ref object i)))
          #+sb-thread
          (when (typep object 'sb-thread:thread)
            (cond ((eq object sb-thread:*current-thread*)
