@@ -127,7 +127,7 @@ then
     exit 1
 fi
 
-# Previously XC host was provided as a positional argument. 
+# Previously XC host was provided as a positional argument.
 if test -n "$legacy_xc_spec"
 then
     SBCL_XC_HOST="$legacy_xc_spec"
@@ -242,8 +242,7 @@ export DEVNULL
 . ./find-gnumake.sh
 find_gnumake
 
-. ./generate-version.sh
-generate_version
+./generate-version.sh
 
 # Now that we've done our option parsing and found various
 # dependencies, write them out to a file to be sourced by other
@@ -301,7 +300,7 @@ case `uname` in
     SunOS)
         sbcl_os="sunos"
         ;;
-    CYGWIN* | WindowsNT | MINGW*)
+    CYGWIN* | WindowsNT | MINGW* | MSYS*)
         sbcl_os="win32"
         ;;
     HP-UX)
@@ -418,6 +417,14 @@ then
             echo "No threads on this platform."
             ;;
     esac
+else
+    case $sbcl_arch in
+        x86|x86-64)
+            case $sbcl_os in
+                linux)
+                    WITH_FEATURES="$WITH_FEATURES :sb-thread"
+            esac
+    esac
 fi
 
 ltf=`pwd`/local-target-features.lisp-expr
@@ -425,8 +432,8 @@ echo //initializing $ltf
 echo ';;;; This is a machine-generated file.' > $ltf
 echo ';;;; Please do not edit it by hand.' >> $ltf
 echo ';;;; See make-config.sh.' >> $ltf
-echo "(lambda (features) (union (set-difference features (list$WITHOUT_FEATURES))" >> $ltf
-printf " (union (list$WITH_FEATURES) (list " >> $ltf
+echo "(lambda (features) (union (set-difference (union features (list$WITH_FEATURES)) (list$WITHOUT_FEATURES))" >> $ltf
+printf " (list " >> $ltf
 
 printf ":%s" "$sbcl_arch" >> $ltf
 
@@ -461,7 +468,7 @@ case "$sbcl_os" in
 		printf ' :largefile' >> $ltf
 		;;
             x86 | x86-64)
-		printf ' :sb-thread :sb-futex :largefile' >> $ltf
+		printf ' :sb-futex :largefile' >> $ltf
 		;;
             ppc)
 		printf ' :sb-futex' >> $ltf
@@ -619,7 +626,7 @@ if [ "$sbcl_arch" = "x86" ]; then
     printf ' :compare-and-swap-vops :unwind-to-frame-and-call-vop :raw-instance-init-vops' >> $ltf
     printf ' :stack-allocatable-closures :stack-allocatable-vectors' >> $ltf
     printf ' :stack-allocatable-lists :stack-allocatable-fixed-objects' >> $ltf
-    printf ' :alien-callbacks :cycle-counter :inline-constants ' >> $ltf
+    printf ' :alien-callbacks :cycle-counter :inline-constants :precise-arg-count-error' >> $ltf
     printf ' :memory-barrier-vops :multiply-high-vops :ash-right-vops :symbol-info-vops' >> $ltf
     case "$sbcl_os" in
     linux | freebsd | gnu-kfreebsd | netbsd | openbsd | sunos | darwin | win32 | dragonfly)
@@ -637,10 +644,11 @@ if [ "$sbcl_arch" = "x86" ]; then
 elif [ "$sbcl_arch" = "x86-64" ]; then
     printf ' :gencgc :stack-grows-downward-not-upward :c-stack-is-control-stack :linkage-table' >> $ltf
     printf ' :compare-and-swap-vops :unwind-to-frame-and-call-vop :raw-instance-init-vops' >> $ltf
+    printf ' :interleaved-raw-slots :precise-arg-count-error' >> $ltf
     printf ' :stack-allocatable-closures :stack-allocatable-vectors' >> $ltf
     printf ' :stack-allocatable-lists :stack-allocatable-fixed-objects' >> $ltf
     printf ' :alien-callbacks :cycle-counter :complex-float-vops' >> $ltf
-    printf ' :float-eql-vops :inline-constants :memory-barrier-vops' >> $ltf
+    printf ' :float-eql-vops :integer-eql-vop :inline-constants :memory-barrier-vops' >> $ltf
     printf ' :multiply-high-vops :sb-simd-pack :ash-right-vops :symbol-info-vops' >> $ltf
 elif [ "$sbcl_arch" = "mips" ]; then
     printf ' :cheneygc :linkage-table' >> $ltf
@@ -708,6 +716,7 @@ elif [ "$sbcl_arch" = "arm" ]; then
     printf ' :ash-right-vops :multiply-high-vops :symbol-info-vops' >> $ltf
     printf ' :stack-allocatable-lists :stack-allocatable-fixed-objects' >> $ltf
     printf ' :stack-allocatable-vectors :stack-allocatable-closures' >> $ltf
+    printf ' :precise-arg-count-error :unwind-to-frame-and-call-vop' >> $ltf
 else
     # Nothing need be done in this case, but sh syntax wants a placeholder.
     echo > /dev/null
@@ -724,7 +733,7 @@ export sbcl_os sbcl_arch
 sh tools-for-build/grovel-features.sh >> $ltf
 
 echo //finishing $ltf
-echo '))))' >> $ltf
+echo ')))' >> $ltf
 
 # FIXME: The version system should probably be redone along these lines:
 #
