@@ -104,7 +104,8 @@
          (push (cons ',name ,name) *entry-points*)
          ,@code
          ,@(generate-return-sequence
-            (or (cadr (assoc :return-style options)) :raw)))
+            (or (cadr (assoc :return-style options)) :raw))
+         (emit-alignment sb!vm:n-lowtag-bits))
        (when sb!xc:*compile-print*
          (format *error-output* "~S assembled~%" ',name)))))
 
@@ -125,9 +126,8 @@
          (vop (make-symbol "VOP")))
     (unless (member return-style '(:raw :full-call :none))
       (error "unknown return-style for ~S: ~S" name return-style))
-    (multiple-value-bind
-        (call-sequence call-temps)
-        (generate-call-sequence name return-style vop)
+    (multiple-value-bind (call-sequence call-temps)
+        (generate-call-sequence name return-style vop options)
       `(define-vop ,(if (atom name) (list name) name)
          (:args ,@(mapcar #'arg-or-res-spec args))
          ,@(let ((index -1))
@@ -145,7 +145,6 @@
                                    :to (:eval 3))
                                   ,(reg-spec-name temp)))
                    temps)
-         ,@call-temps
          (:vop-var ,vop)
          ,@(let ((index -1))
              (mapcar (lambda (res)
@@ -157,13 +156,16 @@
                                     ,(reg-spec-temp res)))
                      results))
          (:results ,@(mapcar #'arg-or-res-spec results))
-         (:ignore ,@(mapcar #'reg-spec-name temps)
-                  ,@(apply #'append
-                           (mapcar #'cdr
-                                   (remove :ignore call-temps
-                                           :test #'neq :key #'car))))
+         ;; This formerly unioned in the contents of an :ignore clause from
+         ;; the value of the 'call-temps' variable, for no good reason afaict.
+         (:ignore ,@(set-difference (mapcar #'reg-spec-name temps)
+                                    (cdr (assoc :call-temps options))))
+         ,@call-temps
+         ;; This too is a tad sleazy - because of how VOP parsing works,
+         ;; any :SAVE-P specified in options supersedes one from call-temps.
+         ;; It would be wiser to signal an error about duplicate options.
          ,@(remove-if (lambda (x)
-                        (member x '(:return-style :cost)))
+                        (member x '(:return-style :cost :call-temps)))
                       options
                       :key #'car)
          (:generator ,cost
