@@ -273,27 +273,38 @@
                      (handler-case (specifier-type type)
                        (parse-unknown-type ()
                          (return (result sb!vm:simple-vector-widetag))))))
-               (if (union-type-p ctype)
-                   (let ((types (union-type-types ctype)))
-                     (cond ((not (every #'numeric-type-p types))
-                            (result sb!vm:simple-vector-widetag))
-                           ((csubtypep ctype (specifier-type 'integer))
-                            (integer-interval-widetag
-                             (reduce #'min types :key #'numeric-type-low)
-                             (reduce #'max types :key #'numeric-type-high)))
-                           ((csubtypep ctype (specifier-type 'double-float))
-                            (result sb!vm:simple-array-double-float-widetag))
-                           ((csubtypep ctype (specifier-type 'single-float))
-                            (result sb!vm:simple-array-single-float-widetag))
-                           #!+long-float
-                           ((csubtypep ctype (specifier-type 'long-float))
-                            (result sb!vm:simple-array-long-float-widetag))
-                           (t
-                            (result sb!vm:simple-vector-widetag))))
-                   (let ((expansion (type-specifier ctype)))
-                     (if (equal expansion type)
-                         (result sb!vm:simple-vector-widetag)
-                         (%vector-widetag-and-n-bits expansion))))))))))))
+               (typecase ctype
+                 (union-type
+                  (let ((types (union-type-types ctype)))
+                    (cond ((not (every #'numeric-type-p types))
+                           (result sb!vm:simple-vector-widetag))
+                          ((csubtypep ctype (specifier-type 'integer))
+                           (integer-interval-widetag
+                            (reduce #'min types :key #'numeric-type-low)
+                            (reduce #'max types :key #'numeric-type-high)))
+                          ((csubtypep ctype (specifier-type 'double-float))
+                           (result sb!vm:simple-array-double-float-widetag))
+                          ((csubtypep ctype (specifier-type 'single-float))
+                           (result sb!vm:simple-array-single-float-widetag))
+                          #!+long-float
+                          ((csubtypep ctype (specifier-type 'long-float))
+                           (result sb!vm:simple-array-long-float-widetag))
+                          (t
+                           (result sb!vm:simple-vector-widetag)))))
+                 (character-set-type
+                  #!-sb-unicode (result sb!vm:simple-base-string-widetag)
+                  #!+sb-unicode
+                  (if (loop for (start . end)
+                            in (character-set-type-pairs ctype)
+                            always (and (< start base-char-code-limit)
+                                        (< end base-char-code-limit)))
+                      (result sb!vm:simple-base-string-widetag)
+                      (result sb!vm:simple-character-string-widetag)))
+                 (t
+                  (let ((expansion (type-specifier ctype)))
+                    (if (equal expansion type)
+                        (result sb!vm:simple-vector-widetag)
+                        (%vector-widetag-and-n-bits expansion)))))))))))))
 
 (defun %complex-vector-widetag (widetag)
   (macrolet ((make-case ()
@@ -373,8 +384,11 @@
                       adjustable fill-pointer
                       displaced-to displaced-index-offset)
   (declare (ignore element-type))
-  (binding* ((dimensions (if (listp dimensions) dimensions (list dimensions)))
-             (array-rank (length (the list dimensions)))
+  (binding* (((array-rank dimension-0)
+              (if (listp dimensions)
+                  (values (length dimensions)
+                          (if dimensions (car dimensions) 1))
+                  (values 1 dimensions)))
              ((initialize initial-data)
               (validate-array-initargs initial-element-p initial-element
                                        initial-contents-p initial-contents
@@ -383,29 +397,41 @@
                           (not adjustable)
                           (null displaced-to))))
     (declare (type array-rank array-rank))
+    (declare (type index dimension-0))
     (cond ((and displaced-index-offset (null displaced-to))
            (error "can't specify :DISPLACED-INDEX-OFFSET without :DISPLACED-TO"))
           ((and simple (= array-rank 1))
-           ;; it's a (SIMPLE-ARRAY * (*))
-           (let* ((length (car dimensions))
-                  (array (allocate-vector-with-widetag widetag length n-bits)))
-             (declare (type index length))
-             (when initial-element-p
-               (fill array initial-element))
-             (when initial-contents-p
-               (unless (= length (length initial-contents))
-                 (error "There are ~W elements in the :INITIAL-CONTENTS, but ~
-                       the vector length is ~W."
-                        (length initial-contents)
-                        length))
-               (replace array initial-contents))
-             array))
+           (let ((vector ; a (SIMPLE-ARRAY * (*))
+                  (allocate-vector-with-widetag widetag dimension-0 n-bits)))
+             ;; presence of at most one :INITIAL-thing keyword was ensured above
+             (cond (initial-element-p
+                    (fill vector initial-element))
+                   (initial-contents-p
+                    (let ((content-length (length initial-contents)))
+                      (unless (= dimension-0 content-length)
+                        (error "There are ~W elements in the :INITIAL-CONTENTS, but ~
+                                the vector length is ~W."
+                               content-length dimension-0)))
+                    (replace vector initial-contents)))
+             vector))
           ((and (arrayp displaced-to)
                 (/= (array-underlying-widetag displaced-to) widetag))
            (error "Array element type of :DISPLACED-TO array does not match specified element type"))
           (t
-           ;; it's either a complex array or a multidimensional array.
-           (let* ((total-size (reduce #'* dimensions))
+           ;; it's non-simple or multidimensional, or both.
+           (when fill-pointer
+             (unless (= array-rank 1)
+               (error "Only vectors can have fill pointers."))
+             (when (and (integerp fill-pointer) (> fill-pointer dimension-0))
+               ;; FIXME: should be TYPE-ERROR?
+               (error "invalid fill-pointer ~W" fill-pointer)))
+           (let* ((total-size
+                   (if (consp dimensions)
+                       (the index (reduce (lambda (a b) (* a (the index b)))
+                                          dimensions))
+                       ;; () is considered to have dimension-0 = 1.
+                       ;; It avoids the REDUCE lambda being called with no args.
+                       dimension-0))
                   (data (or displaced-to
                             (data-vector-from-inits
                              dimensions total-size nil widetag n-bits
@@ -416,25 +442,16 @@
                                 (simple sb!vm:simple-array-widetag)
                                 (t sb!vm:complex-array-widetag))
                           array-rank)))
-             (cond (fill-pointer
-                    (unless (= array-rank 1)
-                      (error "Only vectors can have fill pointers."))
-                    (let ((length (car dimensions)))
-                      (declare (fixnum length))
-                      (setf (%array-fill-pointer array)
-                            (cond ((eq fill-pointer t)
-                                   length)
-                                  ((not (<= fill-pointer length))
-                                     ;; FIXME: should be TYPE-ERROR?
-                                     (error "invalid fill-pointer ~W"
-                                            fill-pointer))
-                                  (t
-                                   fill-pointer))))
-                    (setf (%array-fill-pointer-p array) t))
-                   (t
-                    (setf (%array-fill-pointer array) total-size)
-                    (setf (%array-fill-pointer-p array) nil)))
+             (if fill-pointer
+                 (setf (%array-fill-pointer-p array) t
+                       (%array-fill-pointer array)
+                       (if (eq fill-pointer t) dimension-0 fill-pointer))
+                 (setf (%array-fill-pointer-p array) nil
+                       (%array-fill-pointer array) total-size))
              (setf (%array-available-elements array) total-size)
+             ;; Terrible name for this slot - we displace to the
+             ;; target array's header, if any, not the "ultimate"
+             ;; vector in the chain of displacements.
              (setf (%array-data-vector array) data)
              (setf (%array-displaced-from array) nil)
              (cond (displaced-to
@@ -447,10 +464,11 @@
                       (%save-displaced-array-backpointer array data)))
                    (t
                     (setf (%array-displaced-p array) nil)))
-             (let ((axis 0))
-               (dolist (dim dimensions)
-                 (setf (%array-dimension array axis) dim)
-                 (incf axis)))
+             (if (listp dimensions)
+                 (let ((dims dimensions)) ; avoid "prevents use of assertion"
+                   (dotimes (axis array-rank)
+                     (setf (%array-dimension array axis) (pop dims))))
+                 (setf (%array-dimension array 0) dimension-0))
              array)))))
 
 (defun make-array (dimensions &rest args
@@ -528,15 +546,23 @@ of specialized arrays is supported."
                   (allocate-vector-with-widetag widetag total-size n-bits)
                   (make-array total-size :element-type element-type))))
     (ecase initialize
-     (:initial-element (fill (the vector data) initial-data))
-     (:initial-contents (fill-data-vector data dimensions initial-data))
+     (:initial-element
+      (fill (the vector data) initial-data))
+     (:initial-contents
+      ;; DIMENSIONS can be supplied as a list or integer now
+      (dx-let ((list-of-dims (list dimensions))) ; ok if already a list
+        (fill-data-vector data
+                          (if (listp dimensions) dimensions list-of-dims)
+                          initial-data)))
      ((nil)))
     data))
 
 (defun vector (&rest objects)
   #!+sb-doc
   "Construct a SIMPLE-VECTOR from the given objects."
-  (coerce (the list objects) 'simple-vector))
+  (let ((v (make-array (length objects))))
+    (do-rest-arg ((x i) objects 0 v)
+      (setf (aref v i) x))))
 
 
 ;;;; accessor/setter functions
@@ -795,7 +821,7 @@ of specialized arrays is supported."
 ;;; (setf aref/bit/sbit) are implemented using setf-functions,
 ;;; because they have to work with (setf (apply #'aref array subscripts))
 ;;; All other setfs can be done using setf-functions too, but I
-;;; haven't found technical advantages or disatvantages for either
+;;; haven't found technical advantages or disadvantages for either
 ;;; scheme.
 (defun (setf aref) (new-value array &rest subscripts)
   (declare (truly-dynamic-extent subscripts)
@@ -828,12 +854,14 @@ of specialized arrays is supported."
   #!+sb-doc
   "Return the bit from the BIT-ARRAY at the specified SUBSCRIPTS."
   (declare (type (array bit) bit-array)
+           (truly-dynamic-extent subscripts)
            (optimize (safety 1)))
   (row-major-aref bit-array (%array-row-major-index bit-array subscripts)))
 
 (defun (setf bit) (new-value bit-array &rest subscripts)
   (declare (type (array bit) bit-array)
            (type bit new-value)
+           (truly-dynamic-extent subscripts)
            (optimize (safety 1)))
   (setf (row-major-aref bit-array
                         (%array-row-major-index bit-array subscripts))
@@ -843,6 +871,7 @@ of specialized arrays is supported."
   #!+sb-doc
   "Return the bit from SIMPLE-BIT-ARRAY at the specified SUBSCRIPTS."
   (declare (type (simple-array bit) simple-bit-array)
+           (truly-dynamic-extent subscripts)
            (optimize (safety 1)))
   (row-major-aref simple-bit-array
                   (%array-row-major-index simple-bit-array subscripts)))
@@ -850,6 +879,7 @@ of specialized arrays is supported."
 (defun (setf sbit) (new-value bit-array &rest subscripts)
   (declare (type (simple-array bit) bit-array)
            (type bit new-value)
+           (truly-dynamic-extent subscripts)
            (optimize (safety 1)))
   (setf (row-major-aref bit-array
                         (%array-row-major-index bit-array subscripts))
@@ -1164,7 +1194,8 @@ of specialized arrays is supported."
                  (declare (ignore old-end))
                  (let ((new-data (if (or (and (array-header-p array)
                                               (%array-displaced-p array))
-                                         (> new-length old-length))
+                                         (> new-length old-length)
+                                         (not (adjustable-array-p array)))
                                      (data-vector-from-inits
                                       dimensions new-length
                                       element-type
