@@ -25,21 +25,60 @@
 ;;; FIXME: Are there other tables that need to have entries removed?
 ;;; What about symbols of the form DEF!FOO?
 (defun !unintern-init-only-stuff ()
-  ;; FIXME: I can't see how this wouldn't converge in 1 pass to a fixed point
-  (do ((any-changes? nil nil))
-      (nil)
-    (dolist (package (list-all-packages))
-      (do-symbols (symbol package)
-        (let ((name (symbol-name symbol)))
-          (when (or (string= name "!" :end1 1 :end2 1)
-                    (and (>= (length name) 2)
-                         (string= name "*!" :end1 2 :end2 2)))
-            (/show0 "uninterning cold-init-only symbol..")
-            (/primitive-print name)
-            (unintern symbol package)
-            (setf any-changes? t)))))
-    (unless any-changes?
-      (return))))
+  (flet ((uninternable-p (symbol)
+           (let ((name (symbol-name symbol)))
+             (or (and (>= (length name) 1) (char= (char name 0) #\!))
+                 (and (>= (length name) 2) (string= name "*!" :end1 2))
+                 (memq symbol
+                       ;; DEF!METHOD need no longer be accessible,
+                       ;; but *DELAYED-DEF!METHOD-ARGS* remains,
+                       ;; due to a reference from pcl/methods.lisp.
+                       ;; It would be nice to fix that.
+                       '(def!method
+                         sb!c::sb!pcl sb!c::sb!impl sb!c::sb!kernel
+                         sb!c::sb!c sb!c::sb!int))))))
+    ;; A structure constructor name, in particular !MAKE-SAETP,
+    ;; can't be uninterned if referenced by a defstruct-description.
+    ;; So loop over all structure classoids and clobber any
+    ;; symbol that should be uninternable.
+    (maphash (lambda (classoid layout)
+               (when (structure-classoid-p classoid)
+                 (let ((dd (layout-info layout)))
+                   (setf (dd-constructors dd)
+                         (delete-if (lambda (x)
+                                      (and (consp x) (uninternable-p (car x))))
+                                    (dd-constructors dd))))))
+             (classoid-subclasses (find-classoid t)))
+    ;; Todo: perform one pass, then a full GC, then a final pass to confirm
+    ;; it worked. It shoud be an error if any uninternable symbols remain,
+    ;; but at present there are about 13 other "!" symbols with referers.
+    (with-package-iterator (iter (list-all-packages) :internal :external)
+      (loop (multiple-value-bind (winp symbol accessibility package) (iter)
+              (declare (ignore accessibility))
+              (unless winp
+                (return))
+              (when (uninternable-p symbol)
+                ;; Uninternable symbols which are referenced by other stuff
+                ;; can't disappear from the image, but we don't need to preserve
+                ;; their functions, so FMAKUNBOUND them. This doesn't have
+                ;; the intended effect if the function shares a code-component
+                ;; with non-cold-init lambdas, such as in !CONSTANTP-COLD-INIT
+                ;; and !GLOBALDB-COLD-INIT. Though the cold-init function is
+                ;; never called post-build, it is not discarded. Also, I suspect
+                ;; that the following loop should print nothing, but it does:
+#|
+                (sb-vm::map-allocated-objects
+                  (lambda (obj type size)
+                    (declare (ignore size))
+                    (when (= type sb-vm:code-header-widetag)
+                      (let ((name (sb-c::debug-info-name
+                                   (sb-kernel:%code-debug-info obj))))
+                        (when (and (stringp name) (search "COLD-INIT-FORMS" name))
+                          (print obj)))))
+                  :dynamic)
+|#
+                (fmakunbound symbol)
+                (unintern symbol package)))))))
 
 ;;;; putting ourselves out of our misery when things become too much to bear
 
@@ -62,6 +101,8 @@
 
 ;;; a list of toplevel things set by GENESIS
 (defvar *!reversed-cold-toplevels*)
+(!defvar *!reversed-cold-setf-macros* nil) ; just SETF macros
+(!defvar *!reversed-cold-defuns* nil)      ; just DEFUNs
 
 ;;; a SIMPLE-VECTOR set by GENESIS
 (defvar *!load-time-values*)
@@ -79,9 +120,37 @@
 (defun !cold-init ()
   #!+sb-doc "Give the world a shove and hope it spins."
 
+  #!+sb-show
+  (sb!int::cannot-/show "Test of CANNOT-/SHOW [don't worry - this is expected]")
   (/show0 "entering !COLD-INIT")
+  (setq *readtable* (make-readtable)
+        *previous-case* nil
+        *previous-readtable-case* nil
+        *print-length* 6 *print-level* 3)
+  #!-win32
+  (write-string "COLD-INIT... "
+                (setq *error-output* (!make-cold-stderr-stream)
+                      *standard-output* *error-output*
+                      *trace-output* *error-output*))
 
+  ;; Assert that FBOUNDP doesn't choke when its answer is NIL.
+  ;; It was fine if T because in that case the legality of the arg is certain.
+  ;; And be extra paranoid - ensure that it really gets called.
+  (locally (declare (notinline fboundp)) (fboundp '(setf !zzzzzz)))
+
+  ;; Putting data in a synchronized hashtable (*PACKAGE-NAMES*)
+  ;; requires that the main thread be properly initialized.
   (show-and-call thread-init-or-reinit)
+  ;; Printing of symbols requires that packages be filled in, because
+  ;; OUTPUT-SYMBOL calls FIND-SYMBOL to determine accessibility.
+  (show-and-call !package-cold-init)
+  ;; Fill in the printer's character attribute tables now.
+  ;; If Genesis could write constant arrays into a target core,
+  ;; that would be nice, and would tidy up some other things too.
+  (show-and-call !printer-cold-init)
+  #!-win32
+  (progn (prin1 `(package = ,(package-name *package*)))
+         (terpri))
 
   ;; Anyone might call RANDOM to initialize a hash value or something;
   ;; and there's nothing which needs to be initialized in order for
@@ -93,8 +162,7 @@
 
   (show-and-call !character-database-cold-init)
   (show-and-call !character-name-database-cold-init)
-
-  (show-and-call !package-cold-init)
+  (show-and-call sb!unicode::!unicode-properties-cold-init)
 
   ;; All sorts of things need INFO and/or (SETF INFO).
   (/show0 "about to SHOW-AND-CALL !GLOBALDB-COLD-INIT")
@@ -109,9 +177,10 @@
   (show-and-call !classes-cold-init)
   (show-and-call !early-type-cold-init)
   (show-and-call !late-type-cold-init)
+  ;; See comment at the DEFUN explaining why there are 2 of them.
+  (show-and-call sb!kernel::!late-type-cold-init2)
   (show-and-call !alien-type-cold-init)
   (show-and-call !target-type-cold-init)
-  (show-and-call !vm-type-cold-init)
   ;; FIXME: It would be tidy to make sure that that these cold init
   ;; functions are called in the same relative order as the toplevel
   ;; forms of the corresponding source files.
@@ -125,31 +194,32 @@
   ;; to the subclasses of STRUCTURE-OBJECT.
   (show-and-call sb!kernel::!set-up-structure-object-class)
 
+  (dolist (x (nreverse *!reversed-cold-setf-macros*))
+    (apply #'!quietly-assign-setf-macro x))
+  (dolist (x (nreverse *!reversed-cold-defuns*))
+    (destructuring-bind (name &optional docstring . inline-expansion) x
+      (!%quietly-defun name docstring inline-expansion)))
+
   ;; KLUDGE: Why are fixups mixed up with toplevel forms? Couldn't
   ;; fixups be done separately? Wouldn't that be clearer and better?
   ;; -- WHN 19991204
   (/show0 "doing cold toplevel forms and fixups")
   (/show0 "(LISTP *!REVERSED-COLD-TOPLEVELS*)=..")
   (/hexstr (if (listp *!reversed-cold-toplevels*) "true" "NIL"))
-  (/show0 "about to calculate (LENGTH *!REVERSED-COLD-TOPLEVELS*)")
-  (/show0 "(LENGTH *!REVERSED-COLD-TOPLEVELS*)=..")
-  #!+sb-show (let ((r-c-tl-length (length *!reversed-cold-toplevels*)))
-               (/show0 "(length calculated..)")
-               (let ((hexstr (hexstr r-c-tl-length)))
-                 (/show0 "(hexstr calculated..)")
-                 (/primitive-print hexstr)))
-  (let (#!+sb-show (index-in-cold-toplevels 0)
-        (really-note-if-setf-fun-and-macro #'sb!c::note-if-setf-fun-and-macro)
-        (really-assign-setf-macro #'sb!impl::assign-setf-macro)
-        (really-defun #'sb!impl::%defun))
+  #!-win32
+  (progn (write `("Length(TLFs)= " ,(length *!reversed-cold-toplevels*)))
+         (terpri))
+  #!+win32
+  (progn (/show0 "about to calculate (LENGTH *!REVERSED-COLD-TOPLEVELS*)")
+         (/show0 "(LENGTH *!REVERSED-COLD-TOPLEVELS*)=..")
+         #!+sb-show (let ((r-c-tl-length (length *!reversed-cold-toplevels*)))
+                      (/show0 "(length calculated..)")
+                      (let ((hexstr (hexstr r-c-tl-length)))
+                        (/show0 "(hexstr calculated..)")
+                        (/primitive-print hexstr))))
+  (let (#!+sb-show (index-in-cold-toplevels 0))
     #!+sb-show (declare (type fixnum index-in-cold-toplevels))
 
-    (setf (symbol-function 'sb!c::note-if-setf-fun-and-macro)
-          (lambda (name) (declare (ignore name)) (values))
-          (symbol-function 'sb!impl::assign-setf-macro)
-          #'sb!impl::!quietly-assign-setf-macro
-          (symbol-function 'sb!impl::%defun) #'sb!impl::!%quietly-defun)
-
     (dolist (toplevel-thing (prog1
                                 (nreverse *!reversed-cold-toplevels*)
                               ;; (Now that we've NREVERSEd it, it's
@@ -176,13 +246,11 @@
                                 (third toplevel-thing))
                   (get-lisp-obj-address
                    (svref *!load-time-values* (fourth toplevel-thing)))))
+           (defstruct
+            (apply 'sb!kernel::%defstruct (cdr toplevel-thing)))
            (t
             (!cold-lose "bogus fixup code in *!REVERSED-COLD-TOPLEVELS*"))))
-        (t (!cold-lose "bogus function in *!REVERSED-COLD-TOPLEVELS*"))))
-    (setf (symbol-function 'sb!c::note-if-setf-fun-and-macro)
-          really-note-if-setf-fun-and-macro
-          (symbol-function 'sb!impl::assign-setf-macro) really-assign-setf-macro
-          (symbol-function 'sb!impl::%defun) really-defun))
+        (t (!cold-lose "bogus operation in *!REVERSED-COLD-TOPLEVELS*")))))
   (/show0 "done with loop over cold toplevel forms and fixups")
 
   ;; Set sane values again, so that the user sees sane values instead
@@ -209,7 +277,7 @@
   #!-(and win32 (not sb-thread))
   (show-and-call signal-cold-init-or-reinit)
   (/show0 "enabling internal errors")
-  (setf (extern-alien "internal_errors_enabled" boolean) t)
+  (setf (extern-alien "internal_errors_enabled" int) 1)
 
   (show-and-call float-cold-init-or-reinit)
 
@@ -229,12 +297,12 @@
   (setf *readtable* (copy-readtable *standard-readtable*))
   (setf sb!debug:*debug-readtable* (copy-readtable *standard-readtable*))
   (sb!pretty:!pprint-cold-init)
+  (setq *print-level* nil *print-length* nil) ; restore defaults
 
   ;; the ANSI-specified initial value of *PACKAGE*
   (setf *package* (find-package "COMMON-LISP-USER"))
 
-  (/show0 "done initializing, setting *COLD-INIT-COMPLETE-P*")
-  (setf *cold-init-complete-p* t)
+  (!enable-infinite-error-protector)
 
   ; hppa heap is segmented, lisp and c uses a stub to call eachother
   #!+hpux (%primitive sb!vm::setup-return-from-lisp-stub)
@@ -331,7 +399,7 @@ process to continue normally."
     (thread-init-or-reinit)
     #!-(and win32 (not sb-thread))
     (signal-cold-init-or-reinit)
-    (setf (extern-alien "internal_errors_enabled" boolean) t)
+    (setf (extern-alien "internal_errors_enabled" int) 1)
     (float-cold-init-or-reinit))
   (gc-reinit)
   (foreign-reinit)
@@ -368,6 +436,7 @@ process to continue normally."
               addr (ash addr -4))))
     str))
 
+;; But: you almost never need this. Just use WRITE in all its glory.
 #!+sb-show
 (defun cold-print (x)
   (labels ((%cold-print (obj depthoid)
