@@ -75,7 +75,7 @@
          (limit (truncate (- *print-right-margin* reserve) columns)))
     (flet ((trunc (&optional end)
              (let ((line-end (- limit 2)))
-               (with-output-to-string (s)
+               (with-simple-output-to-string (s)
                  (write-string line s :end (if end
                                                (min end line-end)
                                                line-end))
@@ -350,10 +350,9 @@
         (pprint-indent :block 2 stream)
         (describe-documentation symbol 'type stream (eq t fun))
         (unless (eq t fun)
-          (describe-lambda-list (if (eq :primitive kind)
-                                    (%fun-lambda-list fun)
-                                    (info :type :lambda-list symbol))
-                                stream)
+          ;; even though :translator can store a CTYPE, this is safe
+          ;; because a symbol can't have a non-FUNCTIONP translator.
+          (describe-lambda-list (%fun-lambda-list fun) stream)
           (multiple-value-bind (expansion ok)
               (handler-case (typexpand-1 symbol)
                 (error () (values nil nil)))
@@ -589,7 +588,9 @@
         (multiple-value-bind (fun what lambda-list derived-type declared-type
                               inline methods)
             (cond ((and (not function) (symbolp name) (special-operator-p name))
-                   (let ((fun (symbol-function name)))
+                   ;; The function in the symbol is irrelevant.
+                   ;; Use the def-ir1-translator function for source location.
+                   (let ((fun (info :function :ir1-convert name)))
                      (values fun "a special operator" (%fun-lambda-list fun))))
                   ((and (not function) (symbolp name) (macro-function name))
                    (let ((fun (macro-function name)))
