@@ -173,6 +173,7 @@
   "Return the 1st object in a list or NIL if the list is empty."
   (car list))
 (defun second (list)
+  #!+sb-doc
   "Return the 2nd object in a list or NIL if there is no 2nd object."
   (cadr list))
 (defun third (list)
@@ -484,34 +485,29 @@
 (defun nconc (&rest lists)
    #!+sb-doc
    "Concatenates the lists given as arguments (by changing them)"
-   (declare (truly-dynamic-extent lists) (optimize speed))
+   (declare (optimize speed))
    (flet ((fail (object)
             (error 'type-error
                    :datum object
                    :expected-type 'list)))
-     (do ((top lists (cdr top)))
-         ((null top) nil)
-       (let ((top-of-top (car top)))
-         (typecase top-of-top
+     (do-rest-arg ((result index) lists)
+         (typecase result
            (cons
-            (let* ((result top-of-top)
-                   (splice result))
-              (do ((elements (cdr top) (cdr elements)))
-                  ((endp elements))
-                (let ((ele (car elements)))
+            (let ((splice result))
+              (do-rest-arg ((ele index) lists (1+ index))
                   (typecase ele
                     (cons (rplacd (last splice) ele)
                           (setf splice ele))
                     (null (rplacd (last splice) nil))
-                    (atom (if (cdr elements)
+                    (atom (if (< (1+ index) (length lists))
                               (fail ele)
-                              (rplacd (last splice) ele))))))
+                              (rplacd (last splice) ele)))))
               (return result)))
            (null)
            (atom
-            (if (cdr top)
-                (fail top-of-top)
-                (return top-of-top))))))))
+            (if (< (1+ index) (length lists))
+                (fail result)
+                (return result)))))))
 
 (defun nreconc (x y)
   #!+sb-doc
@@ -558,6 +554,7 @@
                      list)))))))
 
 (defun ldiff (list object)
+  #!+sb-doc
   "Return a new list, whose elements are those of LIST that appear before
    OBJECT. If OBJECT is not a tail of LIST, a copy of LIST is returned.
    LIST must be a proper list or a dotted list."
