@@ -181,15 +181,19 @@ invoked. In that case it will store into PLACE and start over."
         (parse-defmacro lambda-list whole body name 'define-compiler-macro
                         :environment environment)
       (let ((def `(lambda (,whole ,environment)
+                    ,@(sb!c:macro-policy-decls t)
                     ,@local-decs
                     ,body))
             (debug-name (sb!c::debug-name 'compiler-macro-function name)))
-        `(eval-when (:compile-toplevel :load-toplevel :execute)
+        `(progn
+          (eval-when (:compile-toplevel)
+           (sb!c::%compiler-defmacro :compiler-macro-function ',name t))
+          (eval-when (:compile-toplevel :load-toplevel :execute)
            (sb!c::%define-compiler-macro ',name
                                          #',def
                                          ',lambda-list
                                          ,doc
-                                         ',debug-name))))))
+                                         ',debug-name)))))))
 
 ;;; FIXME: This will look remarkably similar to those who have already
 ;;; seen the code for %DEFMACRO in src/code/defmacro.lisp.  Various
@@ -201,6 +205,7 @@ invoked. In that case it will store into PLACE and start over."
                (name definition lambda-list doc debug-name)
              ,@(unless set-p
                  '((declare (ignore lambda-list doc debug-name))))
+             (sb!c::warn-if-compiler-macro-dependency-problem name)
              ;; FIXME: warn about incompatible lambda list with
              ;; respect to parent function?
              (setf (sb!xc:compiler-macro-function name) definition)
@@ -518,7 +523,8 @@ invoked. In that case it will store into PLACE and start over."
   "DECLAIM Declaration*
   Do a declaration or declarations for the global environment."
   `(eval-when (:compile-toplevel :load-toplevel :execute)
-     ,@(mapcar (lambda (spec) `(sb!xc:proclaim ',spec))
+     ,@(mapcar (lambda (spec)
+                 `(sb!c::%proclaim ',spec (sb!c:source-location)))
                specs)))
 
 ;; Avoid unknown return values in emitted code for PRINT-UNREADABLE-OBJECT
@@ -543,3 +549,13 @@ invoked. In that case it will store into PLACE and start over."
   form, or (VALUES NIL the-ERROR-that-was-caught) if an ERROR was handled."
   `(handler-case (progn ,@forms)
      (error (condition) (values nil condition))))
+
+;; A macroexpander helper. Not sure where else to put this.
+(defun funarg-bind/call-forms (funarg arg-forms)
+  (if (typep funarg
+             '(or (cons (eql function) (cons (satisfies legal-fun-name-p) null))
+                  (cons (eql quote) (cons symbol null))))
+      (values nil `(funcall ,funarg . ,arg-forms))
+    (let ((fn-sym (sb!xc:gensym))) ; for ONCE-ONLY-ish purposes
+      (values `((,fn-sym (%coerce-callable-to-fun ,funarg)))
+              `(sb!c::%funcall ,fn-sym . ,arg-forms)))))
