@@ -28,6 +28,14 @@
 
 (defconstant default-line-length 80)
 
+;; We're allowed to DXify the pretty-stream used by PPRINT-LOGICAL-BLOCK.
+;;   "pprint-logical-block and the pretty printing stream it creates have
+;;    dynamic extent. The consequences are undefined if, outside of this
+;;    extent, output is attempted to the pretty printing stream it creates."
+;; However doing that is slightly dangerous since there are a zillion ways
+;; for users to get a hold of the stream and stash it somewhere.
+;; Anyway, just a thought...
+(declaim (maybe-inline make-pretty-stream))
 (defstruct (pretty-stream (:include ansi-stream
                                     (out #'pretty-out)
                                     (sout #'pretty-sout)
@@ -35,13 +43,14 @@
                           (:constructor make-pretty-stream (target))
                           (:copier nil))
   ;; Where the output is going to finally go.
-  (target (missing-arg) :type stream)
+  (target (missing-arg) :type stream :read-only t)
   ;; Line length we should format to. Cached here so we don't have to keep
   ;; extracting it from the target stream.
   (line-length (or *print-right-margin*
                    (sb!impl::line-length target)
                    default-line-length)
-               :type column)
+               :type column
+               :read-only t)
   ;; If non-nil, a function to call before performing OUT or SOUT
   (char-out-oneshot-hook nil :type (or null function))
   ;; A simple string holding all the text that has been output but not yet
@@ -166,6 +175,7 @@
                          (fill-pointer (pretty-stream-buffer-fill-pointer
                                         stream))
                          (new-fill-ptr (+ fill-pointer count)))
+                    (declare (fixnum available count))
                     (if (typep string 'simple-base-string)
                         ;; FIXME: Reimplementing REPLACE, since it
                         ;; can't be inlined and we don't have a
@@ -705,12 +715,9 @@
    from the output and indentation is introduced at the beginning of the
    next line. (See PPRINT-INDENT.)"
   (declare (type (member :linear :miser :fill :mandatory) kind)
-           (type (or stream (member t nil)) stream)
+           (type stream-designator stream)
            (values null))
-  (let ((stream (case stream
-                  ((t) *terminal-io*)
-                  ((nil) *standard-output*)
-                  (t stream))))
+  (let ((stream (out-synonym-of stream)))
     (when (print-pretty-on-stream-p stream)
       (enqueue-newline stream kind)))
   nil)
@@ -732,12 +739,9 @@ The new indentation value does not take effect until the following
 line break."
   (declare (type (member :block :current) relative-to)
            (type real n)
-           (type (or stream (member t nil)) stream)
+           (type stream-designator stream)
            (values null))
-  (let ((stream (case stream
-                  ((t) *terminal-io*)
-                  ((nil) *standard-output*)
-                  (t stream))))
+  (let ((stream (out-synonym-of stream)))
     (when (print-pretty-on-stream-p stream)
       (enqueue-indent stream relative-to (truncate n))))
   nil)
@@ -757,12 +761,9 @@ line break."
        of the current section, not the start of the line."
   (declare (type (member :line :section :line-relative :section-relative) kind)
            (type unsigned-byte colnum colinc)
-           (type (or stream (member t nil)) stream)
+           (type stream-designator stream)
            (values null))
-  (let ((stream (case stream
-                  ((t) *terminal-io*)
-                  ((nil) *standard-output*)
-                  (t stream))))
+  (let ((stream (out-synonym-of stream)))
     (when (print-pretty-on-stream-p stream)
       (enqueue-tab stream kind colnum colinc)))
   nil)
@@ -826,14 +827,14 @@ line break."
 (defvar *standard-pprint-dispatch-table*)
 (defvar *initial-pprint-dispatch-table*)
 
-(defstruct (pprint-dispatch-entry (:copier nil))
+(defstruct (pprint-dispatch-entry (:copier nil) (:predicate nil))
   ;; the type specifier for this entry
   (type (missing-arg) :type t :read-only t)
   ;; a function to test to see whether an object is of this type,
   ;; either (LAMBDA (OBJ) (TYPEP OBJECT TYPE)) or a builtin predicate.
   ;; We don't bother computing this for entries in the CONS
   ;; hash table, because we don't need it.
-  (test-fn nil :type (or function null) :read-only t)
+  (test-fn nil :type (or function null))
   ;; the priority for this guy
   (priority 0 :type real :read-only t)
   ;; T iff one of the original entries.
@@ -848,11 +849,6 @@ line break."
             (pprint-dispatch-entry-priority entry)
             (pprint-dispatch-entry-initial-p entry))))
 
-(defun cons-type-specifier-p (spec)
-  (typep spec '(cons (eql cons)
-                     (cons (cons (member eql member) (cons t null))
-                           null))))
-
 ;; Return T iff E1 is strictly less preferable than E2.
 (defun entry< (e1 e2)
   (declare (type pprint-dispatch-entry e1 e2))
@@ -866,14 +862,14 @@ line break."
           (< (pprint-dispatch-entry-priority e1)
              (pprint-dispatch-entry-priority e2)))))
 
-;; Return the predicate for TYPE. This used to involve rewriting TYPE
-;; into a sexpr due to (CONS <t1> <t2>) not being an official specifier.
-;; Now that it is, there's nothing magical about it.
-(defun compute-test-fn (type function)
+;; Return the predicate for CTYPE, equivalently TYPE-SPEC.
+;; This used to involve rewriting into a sexpr if CONS was involved,
+;; since it was not an official specifier. But now it is.
+(defun compute-test-fn (ctype type-spec function)
   (declare (special sb!c::*backend-type-predicates*))
   ;; Avoid compiling code for an existing structure predicate
-  (or (and (eq (info :type :kind type) :instance)
-           (let ((layout (info :type :compiler-layout type)))
+  (or (and (eq (info :type :kind type-spec) :instance)
+           (let ((layout (info :type :compiler-layout type-spec)))
              (and layout
                   (let ((info (layout-info layout)))
                     (and info
@@ -881,8 +877,7 @@ line break."
                            (and pred (fboundp pred)
                                 (symbol-function pred))))))))
       ;; avoid compiling code for CONS, ARRAY, VECTOR, etc
-      (awhen (assoc (specifier-type type) sb!c::*backend-type-predicates*
-                    :test #'type=)
+      (awhen (assoc ctype sb!c::*backend-type-predicates* :test #'type=)
         (symbol-function (cdr it)))
       ;; OK, compile something
       (let ((name
@@ -890,13 +885,14 @@ line break."
              ;; affects the global environment, when all you want
              ;; is to give the lambda a human-readable label.
              (format nil "~A-P"
-                     (cond ((symbolp type) type)
+                     (cond ((symbolp type-spec) type-spec)
                            ((symbolp function) function)
                            ((%fun-name function))
                            (t
-                            (write-to-string type :pretty nil :escape nil
+                            (write-to-string type-spec :pretty nil :escape nil
                                              :readably nil))))))
-        (compile nil `(named-lambda ,name (object) (typep object ',type))))))
+        (compile nil
+                 `(named-lambda ,name (object) (typep object ',type-spec))))))
 
 (defun copy-pprint-dispatch (&optional (table *print-pprint-dispatch*))
   (declare (type (or pprint-dispatch-table null) table))
@@ -932,6 +928,44 @@ line break."
     (cerror "Frob it anyway!" 'standard-pprint-dispatch-table-modified-error
             :operation operation)))
 
+;; Similar to (NOT CONTAINS-UNKNOWN-TYPE-P), but this is for when you
+;; want to pre-verify that TYPEP won't outright croak, given that you're
+;; going to call it really soon.
+;; Granted, certain checks could pass or fail by short-circuiting,
+;; such as (TYPEP 3 '(OR NUMBER (SATISFIES NO-SUCH-FUN))
+;; but this has to be maximally conservative.
+(defun testable-type-p (ctype)
+  (typecase ctype
+    (unknown-type nil) ; must precede HAIRY because an unknown is HAIRY
+    (hairy-type
+     (let ((spec (hairy-type-specifier ctype)))
+       ;; Anything other than (SATISFIES ...) is testable
+       ;; because there's no reason to suppose that it isn't.
+       (or (neq (car spec) 'satisfies) (fboundp (cadr spec)))))
+    (compound-type (every #'testable-type-p (compound-type-types ctype)))
+    (negation-type (testable-type-p (negation-type-type ctype)))
+    (cons-type (and (testable-type-p (cons-type-car-type ctype))
+                    (testable-type-p (cons-type-cdr-type ctype))))
+    (array-type (testable-type-p (array-type-element-type ctype)))
+    (t t)))
+
+(defun defer-type-checker (entry)
+  (let ((saved-nonce sb!c::*type-cache-nonce*))
+    (lambda (obj)
+      (let ((nonce sb!c::*type-cache-nonce*))
+        (if (eq nonce saved-nonce)
+            nil
+            (let ((ctype (specifier-type (pprint-dispatch-entry-type entry))))
+              (setq saved-nonce nonce)
+              (if (testable-type-p ctype)
+                  (funcall (setf (pprint-dispatch-entry-test-fn entry)
+                                 (compute-test-fn
+                                  ctype
+                                  (pprint-dispatch-entry-type entry)
+                                  (pprint-dispatch-entry-fun entry)))
+                           obj)
+                  nil)))))))
+
 ;; The dispatch mechanism is not quite sophisticated enough to have a guard
 ;; condition on CONS entries. One place this would impact is that you could
 ;; write the full matcher for QUOTE as just a type-specifier. It can be done
@@ -947,12 +981,35 @@ line break."
   (/show0 "entering SET-PPRINT-DISPATCH, TYPE=...")
   (/hexstr type)
   (assert-not-standard-pprint-dispatch-table table 'set-pprint-dispatch)
-  (let* ((consp (cons-type-specifier-p type))
+  (let* ((ctype (or (handler-bind
+                        ((parse-unknown-type
+                          (lambda (c)
+                            (warn "~S is not a recognized type specifier"
+                                  (parse-unknown-type-specifier c)))))
+                      (sb!c::careful-specifier-type type))
+                    (error "~S is not a valid type-specifier" type)))
+         (consp (and (cons-type-p ctype)
+                     (eq (cons-type-cdr-type ctype) *universal-type*)
+                     (member-type-p (cons-type-car-type ctype))
+                     ;; FIXME: (CONS (MEMBER A B ...)) could be hash-based
+                     ;; by expanding it here rather than letting the
+                     ;; compiler see it.
+                     (eql 1 (member-type-size (cons-type-car-type ctype)))))
+         (disabled-p (not (testable-type-p ctype)))
          (entry (if function
                     (make-pprint-dispatch-entry
                      :type type
-                     :test-fn (unless consp (compute-test-fn type function))
+                     :test-fn (unless (or consp disabled-p)
+                                (compute-test-fn ctype type function))
                      :priority priority :fun function))))
+    (when (and function disabled-p)
+      ;; a DISABLED-P test function has to close over the ENTRY
+      (setf (pprint-dispatch-entry-test-fn entry) (defer-type-checker entry))
+      (unless (unknown-type-p ctype) ; already warned in this case
+        ;; But (OR KNOWN UNKNOWN) did not signal - actually it is indeterminate
+        ;; - depending on whather it was cached. I think we should not cache
+        ;; any specifier that contains any unknown anywhere within it.
+        (warn "~S contains an unrecognized type specifier" type)))
     (if consp
         (let ((hashtable (pprint-dispatch-table-cons-entries table))
               (key (second (second type))))
@@ -1161,10 +1218,10 @@ line break."
     (let* ((pretty-p nil)
            (sigil (case (car list)
                     (function "#'")
-                    (quote "'")
                     ;; QUASIQUOTE can't choose not to print prettily.
                     ;; Wrongly nested commas beget unreadable sexprs.
-                    (quasiquote (setq pretty-p t) "`"))))
+                    (quasiquote (setq pretty-p t) "`")
+                    (t "'")))) ; ordinary QUOTE
       (when (or pretty-p *pprint-quote-with-syntactic-sugar*)
         (write-string sigil stream)
         (return-from pprint-quote (output-object (cadr list) stream)))))
@@ -1208,8 +1265,8 @@ line break."
        (write-char #\space stream)
        (pprint-newline :mandatory stream)))))
 
-;;; FIXME: could become SB!XC:DEFMACRO wrapped in EVAL-WHEN (COMPILE EVAL)
-(defmacro pprint-tagbody-guts (stream)
+(eval-when (:compile-toplevel :execute)
+(sb!xc:defmacro pprint-tagbody-guts (stream)
   `(loop
      (pprint-exit-if-list-exhausted)
      (write-char #\space ,stream)
@@ -1218,7 +1275,7 @@ line break."
                       (if (atom form-or-tag) 0 1)
                       ,stream)
        (pprint-newline :linear ,stream)
-       (output-object form-or-tag ,stream))))
+       (output-object form-or-tag ,stream)))))
 
 (defun pprint-tagbody (stream list &rest noise)
   (declare (ignore noise))
@@ -1441,6 +1498,22 @@ line break."
 
 ;;;; the interface seen by regular (ugly) printer and initialization routines
 
+(eval-when (:compile-toplevel :execute)
+(sb!xc:defmacro with-pretty-stream ((stream-var
+                                     &optional (stream-expression stream-var))
+                                    &body body)
+  (let ((flet-name (sb!xc:gensym "WITH-PRETTY-STREAM")))
+    `(flet ((,flet-name (,stream-var)
+              ,@body))
+       (let ((stream ,stream-expression))
+         (if (pretty-stream-p stream)
+             (,flet-name stream)
+             (catch 'line-limit-abbreviation-happened
+               (let ((stream (make-pretty-stream stream)))
+                 (,flet-name stream)
+                 (force-pretty-output stream)))))
+       nil))))
+
 ;;; OUTPUT-PRETTY-OBJECT is called by OUTPUT-OBJECT when
 ;;; *PRINT-PRETTY* is true.
 (defun output-pretty-object (object stream)
@@ -1452,6 +1525,76 @@ line break."
         ;; printing the object after all.
         (output-ugly-object object stream))))
 
+(defun call-logical-block-printer (proc stream prefix per-line-p suffix
+                                   &optional (object nil obj-supplied-p))
+  ;; PREFIX and SUFFIX will be checked for stringness by START-LOGICAL-BLOCK.
+  ;; Doing it here would be more strict, but I really don't think it's worth
+  ;; an extra check. The only observable difference would occur when you have
+  ;; a non-list object which bypasses START-LOGICAL-BLOCK.
+  ;; Also, START-LOGICAL-BLOCK could become an FLET inside here.
+  (declare (function proc))
+  (with-pretty-stream (stream (out-synonym-of stream))
+    (if (not (listp object)) ; implies obj-supplied-p
+        ;; the spec says "If object is not a list, it is printed using WRITE"
+        ;; but I guess this is close enough.
+        (output-object object stream)
+        (dx-let ((state (cons 0 stream)))
+          (if obj-supplied-p
+              (with-circularity-detection (object stream)
+                (descend-into (stream)
+                  (start-logical-block stream prefix per-line-p suffix)
+                  (funcall proc object state stream)
+                  ;; Comment preserved for posterity:
+                  ;;   FIXME: Don't we need UNWIND-PROTECT to ensure this
+                  ;;   always gets executed?
+                  ;; I think not because I wouldn't characterize this as
+                  ;; "cleanup" code. If and only if you follow the accepted
+                  ;; protocol for defining and using print functions should
+                  ;; the behavior be expected to be reasonable and predictable.
+                  ;; Throwing to LINE-LIMIT-ABBREVIATION-HAPPENED is designed
+                  ;; to do the right thing, and printing should not generally
+                  ;; continue to have side-effects if the user felt it necessary
+                  ;; to nonlocally exit in an unexpected way for other reasons.
+                  (end-logical-block stream)))
+              (descend-into (stream)
+                (start-logical-block stream prefix per-line-p suffix)
+                (funcall proc state stream)
+                (end-logical-block stream)))))))
+
+;; Return non-nil if we should keep printing within the logical-block,
+;; or NIL to stop printing due to non-list, length cutoff, or circularity.
+(defun pprint-length-check (obj state)
+  (let ((stream (cdr state)))
+    (cond ((or (not (listp obj))
+               ;; Consider (A . `(,B C)) = (A QUASIQUOTE ,B C)
+               ;; We have to detect this and print as the form on the left,
+               ;; since pretty commas with no containing #\` will be unreadable
+               ;; due to a nesting error.
+               (and (eq (car obj) 'quasiquote) (singleton-p (cdr obj))))
+           (write-string ". " stream)
+           (output-object obj stream)
+           nil)
+          ((and (not *print-readably*) (eql (car state) *print-length*))
+           (write-string "..." stream)
+           nil)
+          ((and obj
+                (plusp (car state))
+                (check-for-circularity obj nil :logical-block))
+           (write-string ". " stream)
+           (output-object obj stream)
+           nil)
+          (t
+           (incf (car state))))))
+
+;; As above, but for logical blocks with an unspecific object.
+(defun pprint-length-check* (state)
+  (let ((stream (cdr state)))
+    (cond ((and (not *print-readably*) (eql (car state) *print-length*))
+           (write-string "..." stream)
+           nil)
+          (t
+           (incf (car state))))))
+
 (defun !pprint-cold-init ()
   (/show0 "entering !PPRINT-COLD-INIT")
   ;; Kludge: We set *STANDARD-PP-D-TABLE* to a new table even though
