@@ -39,10 +39,10 @@
   manual for details."
   (etypecase destination
     (null
-     (with-output-to-string (stream)
+     (with-simple-output-to-string (stream)
        (%format stream control-string format-arguments)))
     (string
-     (with-output-to-string (stream destination)
+     (with-simple-output-to-string (stream destination)
        (%format stream control-string format-arguments)))
     ((member t)
      (%format *standard-output* control-string format-arguments)
@@ -234,11 +234,17 @@
 
 (def-format-interpreter #\C (colonp atsignp params)
   (interpret-bind-defaults () params
-    (if colonp
-        (format-print-named-character (next-arg) stream)
-        (if atsignp
-            (prin1 (next-arg) stream)
-            (write-char (next-arg) stream)))))
+    (let ((arg (next-arg)))
+      (unless (typep arg 'character)
+        (error 'format-error
+               :complaint "~s is not of type CHARACTER."
+               :args (list arg)))
+      (cond (colonp
+             (format-print-named-character arg stream))
+            (atsignp
+             (prin1 arg stream))
+            (t
+             (write-char arg stream))))))
 
 ;;; "printing" as defined in the ANSI CL glossary, which is normative.
 (defun char-printing-p (char)
@@ -303,9 +309,8 @@
          (format-print-integer stream (next-arg) colonp atsignp ,base mincol
                                padchar commachar commainterval))
        (let ((*print-base* ,base)
-             (*print-radix* nil)
-             (*print-escape* nil))
-         (output-object (next-arg) stream))))
+             (*print-radix* nil))
+         (princ (next-arg) stream))))
 ) ; EVAL-WHEN
 
 (def-format-interpreter #\D (colonp atsignp params)
@@ -326,6 +331,11 @@
        (commainterval 3))
       params
     (let ((arg (next-arg)))
+      (unless (or base
+                  (integerp arg))
+        (error 'format-error
+               :complaint "~s is not of type INTEGER."
+               :args (list arg)))
       (if base
           (format-print-integer stream arg colonp atsignp base mincol
                                 padchar commachar commainterval)
@@ -507,7 +517,8 @@
     (number
      (format-write-field stream (decimal-string number) w 1 0 #\space t))
     (t
-     (format-princ stream number nil nil w 1 0 pad))))
+     (let ((*print-base* 10))
+      (format-princ stream number nil nil w 1 0 pad)))))
 
 ;;; We return true if we overflowed, so that ~G can output the overflow char
 ;;; instead of spaces.
@@ -583,7 +594,8 @@
               (format-write-field stream
                                   (decimal-string number)
                                   w 1 0 #\space t)))
-      (format-princ stream number nil nil w 1 0 pad)))
+      (let ((*print-base* 10))
+        (format-princ stream number nil nil w 1 0 pad))))
 
 (defun format-exponent-marker (number)
   (if (typep number *read-default-float-format*)
@@ -678,7 +690,8 @@
               (format-write-field stream
                                   (decimal-string number)
                                   w 1 0 #\space t)))
-      (format-princ stream number nil nil w 1 0 pad)))
+      (let ((*print-base* 10))
+        (format-princ stream number nil nil w 1 0 pad))))
 
 ;;; Raymond Toy writes: same change as for format-exp-aux
 (defun format-general-aux (stream number w d e k ovf pad marker atsign)
@@ -739,9 +752,10 @@
           (dotimes (i (- n pointplace))
             (write-char #\0 stream))
           (write-string str stream)))
-      (format-write-field stream
-                          (decimal-string number)
-                          w 1 0 #\space t)))
+      (let ((*print-base* 10))
+        (format-write-field stream
+                            (princ-to-string number)
+                            w 1 0 #\space t))))
 
 ;;;; FORMAT interpreters and support functions for line/page breaks etc.
 
@@ -1089,7 +1103,7 @@
   (multiple-value-bind (segments first-semi close remaining)
       (parse-format-justification directives)
     (setf args
-          (if (format-directive-colonp close)
+          (if (format-directive-colonp close) ; logical block vs. justification
               (multiple-value-bind (prefix per-line-p insides suffix)
                   (parse-format-logical-block segments colonp first-semi
                                               close params string end)
@@ -1104,6 +1118,16 @@
                          :complaint "~D illegal directive~:P found inside justification block"
                          :args (list count)
                          :references (list '(:ansi-cl :section (22 3 5 2)))))
+                ;; ANSI does not explicitly say that an error should
+                ;; be signalled, but the @ modifier is not explicitly
+                ;; allowed for ~> either.
+                (when (format-directive-atsignp close)
+                  (error 'format-error
+                         :complaint "@ modifier not allowed in close ~
+                         directive of justification ~
+                         block (i.e. ~~<...~~@>."
+                         :offset (1- (format-directive-end close))
+                         :references (list '(:ansi-cl :section (22 3 6 2)))))
                 (interpret-format-justification stream orig-args args
                                                 segments colonp atsignp
                                                 first-semi params))))
@@ -1126,7 +1150,7 @@
                      (len (or (sb!impl::line-length stream) 72)))
                     (format-directive-params first-semi)
                   (setf newline-string
-                        (with-output-to-string (stream)
+                        (with-simple-output-to-string (stream)
                           (setf args
                                 (interpret-directive-list stream
                                                           (pop segments)
@@ -1135,7 +1159,7 @@
                   (setf extra-space extra)
                   (setf line-len len)))
               (dolist (segment segments)
-                (push (with-output-to-string (stream)
+                (push (with-simple-output-to-string (stream)
                         (setf args
                               (interpret-directive-list stream segment
                                                         orig-args args)))
