@@ -267,9 +267,9 @@ created and old ones may exit at any time."
   (setf sb!impl::*exit-lock* (make-mutex :name "Exit Lock")
         *make-thread-lock* (make-mutex :name "Make-Thread Lock"))
   (let ((initial-thread (%make-thread :name "main thread"
-                                      :%alive-p t
-                                      :os-thread (current-thread-os-thread))))
-    (setq *initial-thread* initial-thread
+                                      :%alive-p t)))
+    (setf (thread-os-thread initial-thread) (current-thread-os-thread)
+          *initial-thread* initial-thread
           *current-thread* initial-thread)
     (grab-mutex (thread-result-lock *initial-thread*))
     ;; Either *all-threads* is empty or it contains exactly one thread
@@ -744,8 +744,11 @@ IF-NOT-OWNER is :FORCE)."
 
 ;;;; Waitqueues/condition variables
 
+;; For possible DX-allocating, see the notinline counterpart after the
+;; PROGN below
+(declaim (inline make-waitqueue))
 #!+(or (not sb-thread) sb-futex)
-(defstruct (waitqueue (:constructor %make-waitqueue))
+(defstruct (waitqueue (:constructor make-waitqueue (&key name)))
   #!+sb-doc
   "Waitqueue type."
   (name nil :type (or null thread-name))
@@ -754,7 +757,7 @@ IF-NOT-OWNER is :FORCE)."
 
 #!+(and sb-thread (not sb-futex))
 (progn
-  (defstruct (waitqueue (:constructor %make-waitqueue))
+  (defstruct (waitqueue (:constructor make-waitqueue (&key name)))
     #!+sb-doc
     "Waitqueue type."
     (name nil :type (or null thread-name))
@@ -811,19 +814,17 @@ IF-NOT-OWNER is :FORCE)."
                               (thread-waiting-for next) queue nil))
                (decf n)))
     nil))
+(declaim (notinline make-waitqueue))
 
 (def!method print-object ((waitqueue waitqueue) stream)
   (print-unreadable-object (waitqueue stream :type t :identity t)
     (format stream "~@[~A~]" (waitqueue-name waitqueue))))
 
-(defun make-waitqueue (&key name)
-  #!+sb-doc
-  "Create a waitqueue."
-  (%make-waitqueue :name name))
-
 #!+sb-doc
 (setf (fdocumentation 'waitqueue-name 'function)
-      "The name of the waitqueue. Setfable.")
+      "The name of the waitqueue. Setfable."
+      (fdocumentation 'make-waitqueue 'function)
+      "Create a waitqueue.")
 
 #!+(and sb-thread sb-futex)
 (define-structure-slot-addressor waitqueue-token-address
@@ -1011,7 +1012,9 @@ must be held by this thread during this call."
 
 ;;;; Semaphores
 
-(defstruct (semaphore (:constructor %make-semaphore (name %count)))
+(declaim (inline make-semaphore)) ;; for possible DX-allocating
+(defstruct (semaphore (:constructor make-semaphore
+                          (&key name ((:count %count) 0))))
   #!+sb-doc
   "Semaphore type. The fact that a SEMAPHORE is a STRUCTURE-OBJECT
 should be considered an implementation detail, and may change in the
@@ -1021,10 +1024,13 @@ future."
   (waitcount 0 :type sb!vm:word)
   (mutex (make-mutex))
   (queue (make-waitqueue)))
+(declaim (notinline make-semaphore))
 
 #!+sb-doc
 (setf (fdocumentation 'semaphore-name 'function)
-      "The name of the semaphore INSTANCE. Setfable.")
+      "The name of the semaphore INSTANCE. Setfable."
+      (fdocumentation 'make-semaphore 'function)
+      "Create a semaphore with the supplied COUNT and NAME.")
 
 (defstruct (semaphore-notification (:constructor make-semaphore-notification ())
                                    (:copier nil))
@@ -1063,11 +1069,6 @@ WAIT-ON-SEMAPHORE or TRY-SEMAPHORE."
   (barrier (:read))
   (semaphore-%count instance))
 
-(defun make-semaphore (&key name (count 0))
-  #!+sb-doc
-  "Create a semaphore with the supplied COUNT and NAME."
-  (%make-semaphore name count))
-
 (defun wait-on-semaphore (semaphore &key timeout notification)
   #!+sb-doc
   "Decrement the count of SEMAPHORE if the count would not be negative. Else
@@ -1223,7 +1224,7 @@ on this semaphore, then N of them is woken up."
   ;; Lisp-side cleanup
   (with-all-threads-lock
     (setf (thread-%alive-p thread) nil)
-    (setf (thread-os-thread thread) nil)
+    (setf (thread-os-thread thread) 0)
     (setq *all-threads* (delq thread *all-threads*))
     (when *session*
       (%delete-thread-from-session thread *session*))))
@@ -1383,7 +1384,7 @@ session."
 
 #!+sb-thread
 (defun initial-thread-function-trampoline
-    (thread setup-sem real-function arguments arg1 arg2 arg3)
+    (thread setup-sem real-function thread-list arguments arg1 arg2 arg3)
   ;; In time we'll move some of the binding presently done in C here
   ;; too.
   ;;
@@ -1400,7 +1401,7 @@ session."
   ;; GC. Be very careful.
   (let* ((*current-thread* thread)
          (*restart-clusters* nil)
-         (*handler-clusters* (sb!kernel::initial-handler-clusters))
+         (*handler-clusters* sb!kernel::**initial-handler-clusters**)
          (*exit-in-process* nil)
          (sb!impl::*deadline* nil)
          (sb!impl::*deadline-seconds* nil)
@@ -1410,60 +1411,76 @@ session."
          ;; internal printer variables
          (sb!impl::*previous-case* nil)
          (sb!impl::*previous-readtable-case* nil)
-         (sb!impl::*internal-symbol-output-fun* nil)
+         (sb!impl::*internal-symbol-output-fun* #'error)
          (sb!impl::*descriptor-handlers* nil)) ; serve-event
+    (declare (inline make-restart)) ;; to allow DX-allocation
     ;; Binding from C
     (setf sb!vm:*alloc-signal* *default-alloc-signal*)
     (setf (thread-os-thread thread) (current-thread-os-thread))
     (with-mutex ((thread-result-lock thread))
-      (let* ((cell1 (shiftf (thread-result thread) nil))
-             (cell2 (cdr cell1)))
+      ;; Normally the list is allocated in the parent thread to
+      ;; avoid consing in a nascent thread.
+
+      (let* ((thread-list (or thread-list
+                              (list thread thread)))
+             (session-cons (cdr thread-list)))
         (with-all-threads-lock
-          (setf *all-threads* (rplacd (rplaca cell1 thread) *all-threads*)))
+          (setf (cdr thread-list) *all-threads*
+                *all-threads* thread-list))
         (let ((session *session*))
           (with-session-lock (session)
-            (setf (session-threads session)
-                  (rplacd (rplaca cell2 thread) (session-threads session))))))
+            (setf (cdr session-cons) (session-threads session)
+                  (session-threads session) session-cons))))
       (setf (thread-%alive-p thread) t)
-      (when setup-sem (signal-semaphore setup-sem))
+
+      (when setup-sem
+        (signal-semaphore setup-sem)
+        ;; setup-sem was dx-allocated, set it to NIL so that the
+        ;; backtrace doesn't get confused
+        (setf setup-sem nil))
+
       ;; Using handling-end-of-the-world would be a bit tricky
       ;; due to other catches and interrupts, so we essentially
       ;; re-implement it here. Once and only once more.
       (catch 'sb!impl::toplevel-catcher
         (catch 'sb!impl::%end-of-the-world
           (catch '%abort-thread
-            (with-simple-restart
-                (abort "~@<Abort thread (~A)~@:>" *current-thread*)
+            (restart-bind ((abort
+                             (lambda ()
+                               (throw '%abort-thread nil))
+                             :report-function
+                             (lambda (stream)
+                               (format stream "~@<abort thread (~a)~@:>"
+                                       *current-thread*))))
               (without-interrupts
-                  (unwind-protect
-                       (with-local-interrupts
-                         (setf *gc-inhibit* nil) ;for foreign callbacks
-                         (sb!unix::unblock-deferrable-signals)
-                         (setf (thread-result thread)
-                               (prog1
-                                   (cons t
-                                         (multiple-value-list
-                                          (unwind-protect
-                                               (catch '%return-from-thread
-                                                 (if (listp arguments)
-                                                     (apply real-function arguments)
-                                                     (funcall real-function arg1 arg2 arg3)))
-                                            (when *exit-in-process*
-                                              (sb!impl::call-exit-hooks)))))
-                                 #!+sb-safepoint
-                                 (sb!kernel::gc-safepoint))))
-                    ;; We're going down, can't handle interrupts
-                    ;; sanely anymore. GC remains enabled.
-                    (block-deferrable-signals)
-                    ;; We don't want to run interrupts in a dead
-                    ;; thread when we leave WITHOUT-INTERRUPTS.
-                    ;; This potentially causes important
-                    ;; interupts to be lost: SIGINT comes to
-                    ;; mind.
-                    (setq *interrupt-pending* nil)
-                    #!+sb-thruption
-                    (setq *thruption-pending* nil)
-                    (handle-thread-exit thread)))))))))
+                (unwind-protect
+                     (with-local-interrupts
+                       (setf *gc-inhibit* nil) ;for foreign callbacks
+                       (sb!unix::unblock-deferrable-signals)
+                       (setf (thread-result thread)
+                             (prog1
+                                 (multiple-value-list
+                                  (unwind-protect
+                                       (catch '%return-from-thread
+                                         (if (listp arguments)
+                                             (apply real-function arguments)
+                                             (funcall real-function arg1 arg2 arg3)))
+                                    (when *exit-in-process*
+                                      (sb!impl::call-exit-hooks))))
+                               #!+sb-safepoint
+                               (sb!kernel::gc-safepoint))))
+                  ;; we're going down, can't handle interrupts
+                  ;; sanely anymore. gc remains enabled.
+                  (block-deferrable-signals)
+                  ;; we don't want to run interrupts in a dead
+                  ;; thread when we leave without-interrupts.
+                  ;; this potentially causes important
+                  ;; interupts to be lost: sigint comes to
+                  ;; mind.
+                  (setq *interrupt-pending* nil)
+                  #!+sb-thruption
+                  (setq *thruption-pending* nil)
+                  (handle-thread-exit thread)))))))))
   (values))
 
 (defun make-thread (function &key name arguments ephemeral)
@@ -1486,6 +1503,9 @@ See also: RETURN-FROM-THREAD, ABORT-THREAD."
                        'make-thread arguments)
   #!+sb-thread
   (let ((thread (%make-thread :name name :%ephemeral-p ephemeral)))
+    (declare (inline make-semaphore
+                     make-waitqueue
+                     make-mutex))
     (let* ((setup-sem (make-semaphore :name "Thread setup semaphore"))
            (real-function (coerce function 'function))
            (arguments     (if (listp arguments)
@@ -1494,27 +1514,33 @@ See also: RETURN-FROM-THREAD, ABORT-THREAD."
            #!+win32
            (fp-modes (dpb 0 sb!vm::float-sticky-bits ;; clear accrued bits
                           (sb!vm:floating-point-modes)))
-           (initial-function
-             (named-lambda initial-thread-function ()
-               ;; Win32 doesn't inherit parent thread's FP modes,
-               ;; while it seems to happen everywhere else
-               #!+win32
-               (setf (sb!vm:floating-point-modes) fp-modes)
-               ;; As it is, this lambda must not cons until we are
-               ;; ready to run GC. Be very careful.
-               (initial-thread-function-trampoline
-                thread setup-sem real-function arguments nil nil nil))))
-      ;; If the starting thread is stopped for gc before it signals
-      ;; the semaphore then we'd be stuck.
-      (assert (not *gc-inhibit*))
-      ;; Keep INITIAL-FUNCTION pinned until the child thread is
-      ;; initialized properly. Wrap the whole thing in
-      ;; WITHOUT-INTERRUPTS because we pass INITIAL-FUNCTION to
-      ;; another thread.
-      (with-system-mutex (*make-thread-lock*)
-        (with-pinned-objects (initial-function)
+           ;; Allocate in the parent
+           (thread-list (list thread thread)))
+      (declare (dynamic-extent setup-sem))
+      ;; It could be just pinned instead, but implementing DX closures
+      ;; is much easier than implementing threading.
+      #!-(or stack-allocatable-closures stack-allocatable-fixed-objects)
+      #.(error "Stack allocatable closures required.")
+      (dx-flet ((initial-thread-function ()
+                  ;; Win32 doesn't inherit parent thread's FP modes,
+                  ;; while it seems to happen everywhere else
+                  #!+win32
+                  (setf (sb!vm:floating-point-modes) fp-modes)
+                  ;; As it is, this lambda must not cons until we are
+                  ;; ready to run GC. Be very careful.
+                  (initial-thread-function-trampoline
+                   thread setup-sem real-function thread-list
+                   arguments nil nil nil)))
+        ;; If the starting thread is stopped for gc before it signals
+        ;; the semaphore then we'd be stuck.
+        (assert (not *gc-inhibit*))
+        ;; Keep INITIAL-FUNCTION in the dynamic extent until the child
+        ;; thread is initialized properly. Wrap the whole thing in
+        ;; WITHOUT-INTERRUPTS because we pass INITIAL-FUNCTION to
+        ;; another thread.
+        (with-system-mutex (*make-thread-lock*)
           (if (zerop
-               (%create-thread (get-lisp-obj-address initial-function)))
+               (%create-thread (get-lisp-obj-address #'initial-thread-function)))
               (setf thread nil)
               (wait-on-semaphore setup-sem)))))
     (or thread (error "Could not create a new thread."))))
@@ -1542,9 +1568,9 @@ subject to change."
                      (allow-with-interrupts
                        ;; Don't use the timeout if the thread is not alive anymore.
                        (grab-mutex lock :timeout (and (thread-alive-p thread) timeout))))
-               (cond ((car (thread-result thread))
+               (cond ((listp (thread-result thread))
                       (return-from join-thread
-                        (values-list (cdr (thread-result thread)))))
+                        (values-list (thread-result thread))))
                      (defaultp
                       (return-from join-thread default))
                      (t
@@ -1560,11 +1586,11 @@ subject to change."
   "Deprecated. Same as TERMINATE-THREAD."
   (terminate-thread thread))
 
-#!+sb-safepoint
+#!+sb-thread
 (defun enter-foreign-callback (arg1 arg2 arg3)
   (initial-thread-function-trampoline
    (make-foreign-thread :name "foreign callback")
-   nil #'sb!alien::enter-alien-callback t arg1 arg2 arg3))
+   nil #'sb!alien::enter-alien-callback nil t arg1 arg2 arg3))
 
 (defmacro with-interruptions-lock ((thread) &body body)
   `(with-system-mutex ((thread-interruptions-lock ,thread))
@@ -1792,14 +1818,15 @@ assume that unknown code can safely be terminated using TERMINATE-THREAD."
   ;; Get values from the TLS area of the current thread.
   (defun %thread-local-references ()
     ;; TLS-INDEX-START is a word number relative to thread base.
-    ;; *FREE-TLS-INDEX* - which should only be read by machine code  - is an
+    ;; *FREE-TLS-INDEX* - which is only manipulated by machine code  - is an
     ;; offset from thread base to the next usable TLS cell as a raw word
     ;; manifesting in Lisp as a fixnum. Convert it from byte number to word
     ;; number before using it as the upper bound for the sequential scan.
-    (declare (special sb!vm::*free-tls-index*))
     (loop for index from tls-index-start
-          below (ash (get-lisp-obj-address sb!vm::*free-tls-index*)
-                     (- sb!vm:word-shift))
+          ;; On x86-64, mask off the sign bit. It is used as a semaphore.
+          below (ash (logand #!+x86-64 most-positive-fixnum
+                             sb!vm::*free-tls-index*)
+                     (+ (- sb!vm:word-shift) sb!vm:n-fixnum-tag-bits))
           for obj = (sb!kernel:%make-lisp-obj
                      (sb!sys:sap-int (sb!vm::current-thread-offset-sap index)))
           unless (or (member (widetag-of obj)
@@ -1867,8 +1894,6 @@ mechanism for inter-thread communication."
                  :info (list :write :unbound-in-thread))
           (values nil nil))))
 
-(defun sb!vm::locked-symbol-global-value-add (symbol-name delta)
-  (sb!vm::locked-symbol-global-value-add symbol-name delta))
 
 
 ;;;; Stepping
