@@ -38,6 +38,11 @@
   ;; the EQ table.
   (equal-table (make-hash-table :test 'equal) :type hash-table)
   (eq-table (make-hash-table :test 'eq) :type hash-table)
+  ;; Hashtable mapping a string to a list of fop-table indices of
+  ;; symbols whose name is that string. For any name as compared
+  ;; by STRING= there can be a symbol whose name is a base string
+  ;; and/or a symbol whose name is not a base string.
+  (string=-table (make-hash-table :test 'equal) :type hash-table)
   ;; the table's current free pointer: the next offset to be used
   (table-free 0 :type index)
   ;; an alist (PACKAGE . OFFSET) of the table offsets for each package
@@ -131,6 +136,24 @@
     (dump-byte (logand n #xff) fasl-output))
   (values))
 
+(defun dump-fop+operands (fasl-output opcode arg1 &optional (arg2 0 arg2p))
+  (declare (type (unsigned-byte 8) opcode) (type word arg1 arg2))
+  (multiple-value-bind (modifier1 modifier2)
+      (flet ((fop-opcode-modifier (arg)
+               ;; The compiler should be able to figure that 32-bit builds
+               ;; can not use the 8-byte size, I think.
+               (if (< arg #x10000)
+                   (if (< arg #x100) 0 1)
+                   (if (< arg (ash 1 32)) 2 3))))
+        (declare (inline fop-opcode-modifier))
+        (values (fop-opcode-modifier arg1)
+                (if arg2p (fop-opcode-modifier arg2) 0)))
+    (dump-byte (logior opcode (ash modifier2 2) modifier1) fasl-output)
+    ;; special-case code optimized for each size might improve speed here
+    (dump-integer-as-n-bytes arg1 (ash 1 modifier1) fasl-output)
+    (when arg2p
+      (dump-integer-as-n-bytes arg2 (ash 1 modifier2) fasl-output))))
+
 ;;; Setting this variable to an (UNSIGNED-BYTE 32) value causes
 ;;; DUMP-FOP to use it as a counter and emit a FOP-NOP4 with the
 ;;; counter value before every ordinary fop. This can make it easier
@@ -147,38 +170,33 @@
 ;;;
 ;;; FIXME: Compiler macros, frozen classes, inlining, and similar
 ;;; optimizations should be conditional on #!+SB-FROZEN.
-(defmacro dump-fop (fs file)
-  (let* ((fs (eval fs))
-         (val (get fs 'fop-code)))
-    (if val
+(defmacro dump-fop (fs-expr file &rest args)
+  (let* ((fs (eval fs-expr))
+         (val (or (get fs 'opcode)
+                  (error "compiler bug: ~S is not a legal fasload operator."
+                         fs-expr)))
+         (fop-argc
+          (if (>= val +2-operand-fops+)
+              2
+              (sbit (car **fop-signatures**) (ash val -2)))))
+    (cond
+      ((not (eql (length args) fop-argc))
+       (error "~S takes ~D argument~:P" fs fop-argc))
+      (t
       `(progn
          #!+sb-show
          (when *fop-nop4-count*
-           (dump-byte ,(get 'fop-nop4 'fop-code) ,file)
+           (dump-byte (get 'fop-nop4 'fop-code) ,file)
            (dump-integer-as-n-bytes (mod (incf *fop-nop4-count*) (expt 2 32))
                                     4 ,file))
-         (dump-byte ',val ,file))
-      (error "compiler bug: ~S is not a legal fasload operator." fs))))
-
-;;; Dump a FOP-CODE along with an integer argument, choosing the FOP
-;;; based on whether the argument will fit in a single byte.
-;;;
-;;; FIXME: This, like DUMP-FOP, should be a function with a
-;;; compiler-macro expansion.
-(defmacro dump-fop* (n byte-fop word-fop file)
-  (once-only ((n-n n)
-              (n-file file))
-    `(cond ((< ,n-n 256)
-            (dump-fop ',byte-fop ,n-file)
-            (dump-byte ,n-n ,n-file))
-           (t
-            (dump-fop ',word-fop ,n-file)
-            (dump-word ,n-n ,n-file)))))
+         ,(if (zerop fop-argc)
+              `(dump-byte ,val ,file)
+              `(dump-fop+operands ,file ,val ,@args)))))))
 
 ;;; Push the object at table offset Handle on the fasl stack.
 (defun dump-push (handle fasl-output)
   (declare (type index handle) (type fasl-output fasl-output))
-  (dump-fop* handle fop-byte-push fop-push fasl-output)
+  (dump-fop 'fop-push fasl-output handle)
   (values))
 
 ;;; Pop the object currently on the fasl stack top into the table, and
@@ -531,7 +549,7 @@
      (dump-integer-as-n-bytes n #.sb!vm:n-word-bytes file))
     (t
      (let ((bytes (ceiling (1+ (integer-length n)) 8)))
-       (dump-fop* bytes fop-small-integer fop-integer file)
+       (dump-fop 'fop-integer file bytes)
        (dump-integer-as-n-bytes n bytes file)))))
 
 (defun dump-float (x file)
@@ -616,7 +634,7 @@
   (cond ((cdr (assoc pkg (fasl-output-packages file) :test #'eq)))
         (t
          (let ((s (package-name pkg)))
-           (dump-fop* (length s) fop-small-named-package-save fop-named-package-save file)
+           (dump-fop 'fop-named-package-save file (length s))
            #+sb-xc-host
            (dump-base-chars-of-string (coerce s 'simple-base-string) file)
            #-sb-xc-host
@@ -647,41 +665,46 @@
 (defun dump-list (list file)
   (aver (and list
              (not (gethash list (fasl-output-circularity-table file)))))
-  (do* ((l list (cdr l))
-        (n 0 (1+ n))
-        (circ (fasl-output-circularity-table file)))
-       ((atom l)
-        (cond ((null l)
-               (terminate-undotted-list n file))
-              (t
-               (sub-dump-object l file)
-               (terminate-dotted-list n file))))
-    (declare (type index n))
-    (let ((ref (gethash l circ)))
-      (when ref
-        (push (make-circularity :type :rplacd
-                                :object list
-                                :index (1- n)
-                                :value l
-                                :enclosing-object ref)
-              *circularities-detected*)
-        (terminate-undotted-list n file)
-        (return)))
+  (let ((circ (fasl-output-circularity-table file)))
+   (flet ((cdr-circularity (obj n)
+            (let ((ref (gethash obj circ)))
+              (when ref
+                (push (make-circularity :type :rplacd
+                                        :object list
+                                        :index (1- n)
+                                        :value obj
+                                        :enclosing-object ref)
+                      *circularities-detected*)
+                (terminate-undotted-list n file)
+                t))))
+     (do* ((l list (cdr l))
+           (n 0 (1+ n)))
+          ((atom l)
+           (cond ((null l)
+                  (terminate-undotted-list n file))
+                 (t
+                  (cond ((cdr-circularity l n))
+                        (t
+                         (sub-dump-object l file)
+                         (terminate-dotted-list n file))))))
+       (declare (type index n))
+       (when (cdr-circularity l n)
+         (return))
 
-    (setf (gethash l circ) list)
+       (setf (gethash l circ) list)
 
-    (let* ((obj (car l))
-           (ref (gethash obj circ)))
-      (cond (ref
-             (push (make-circularity :type :rplaca
-                                     :object list
-                                     :index n
-                                     :value obj
-                                     :enclosing-object ref)
-                   *circularities-detected*)
-             (sub-dump-object nil file))
-            (t
-             (sub-dump-object obj file))))))
+       (let* ((obj (car l))
+              (ref (gethash obj circ)))
+         (cond (ref
+                (push (make-circularity :type :rplaca
+                                        :object list
+                                        :index n
+                                        :value obj
+                                        :enclosing-object ref)
+                      *circularities-detected*)
+                (sub-dump-object nil file))
+               (t
+                (sub-dump-object obj file))))))))
 
 (defun terminate-dotted-list (n file)
   (declare (type index n) (type fasl-output file))
@@ -791,7 +814,7 @@
        (length (length v))
        (circ (fasl-output-circularity-table file)))
       ((= index length)
-       (dump-fop* length fop-small-vector fop-vector file))
+       (dump-fop 'fop-vector file length))
     (let* ((obj (aref v index))
            (ref (gethash obj circ)))
       (cond (ref
@@ -890,14 +913,7 @@
 ;;; Dump characters and string-ish things.
 
 (defun dump-character (char file)
-  (let ((code (sb!xc:char-code char)))
-    (cond
-      ((< code 256)
-       (dump-fop 'fop-short-character file)
-       (dump-byte code file))
-      (t
-       (dump-fop 'fop-character file)
-       (dump-word code file)))))
+  (dump-fop 'fop-character file (sb!xc:char-code char)))
 
 (defun dump-base-chars-of-string (s fasl-output)
   (declare #+sb-xc-host (type simple-string s)
@@ -912,7 +928,7 @@
 (defun dump-simple-base-string (s file)
   #+sb-xc-host (declare (type simple-string s))
   #-sb-xc-host (declare (type simple-base-string s))
-  (dump-fop* (length s) fop-small-base-string fop-base-string file)
+  (dump-fop 'fop-base-string file (length s))
   (dump-base-chars-of-string s file)
   (values))
 
@@ -922,6 +938,7 @@
   (declare (type fasl-output file))
   (let* ((pname (symbol-name s))
          (pname-length (length pname))
+         (dumped-as-copy nil)
          (pkg (symbol-package s)))
     ;; see comment in genesis: we need this here for repeatable fasls
     #+sb-xc-host
@@ -934,10 +951,21 @@
         (setq pkg sb!int:*cl-package*)))
 
     (cond ((null pkg)
-           (dump-fop* pname-length
-                      fop-uninterned-small-symbol-save
-                      fop-uninterned-symbol-save
-                      file))
+           (let ((this-base-p (typep pname 'base-string)))
+             (dolist (lookalike (gethash pname (fasl-output-string=-table file))
+                                (dump-fop 'fop-uninterned-symbol-save
+                                          file pname-length))
+               ;; Find the right kind of lookalike symbol.
+               ;; actually this seems pretty bogus - afaict, we don't correctly
+               ;; preserve the type of the string (base or character) anyway,
+               ;; but if we did, then this would be right also.
+               ;; [what about a symbol whose name is a (simple-array nil (0))?]
+               (let ((that-base-p (typep (symbol-name lookalike) 'base-string)))
+                 (when (or (and this-base-p that-base-p)
+                           (and (not this-base-p) (not that-base-p)))
+                   (dump-fop 'fop-copy-symbol-save file
+                             (gethash lookalike (fasl-output-eq-table file)))
+                   (return (setq dumped-as-copy t)))))))
           ;; CMU CL had FOP-SYMBOL-SAVE/FOP-SMALL-SYMBOL-SAVE fops which
           ;; used the current value of *PACKAGE*. Unfortunately that's
           ;; broken w.r.t. ANSI Common Lisp semantics, so those are gone
@@ -947,32 +975,19 @@
           ;;        fop-small-symbol-save
           ;;        fop-symbol-save file))
           ((eq pkg sb!int:*cl-package*)
-           (dump-fop* pname-length
-                      fop-lisp-small-symbol-save
-                      fop-lisp-symbol-save
-                      file))
+           (dump-fop 'fop-lisp-symbol-save file pname-length))
           ((eq pkg sb!int:*keyword-package*)
-           (dump-fop* pname-length
-                      fop-keyword-small-symbol-save
-                      fop-keyword-symbol-save
-                      file))
-          ((< pname-length 256)
-           (dump-fop* (dump-package pkg file)
-                      fop-small-symbol-in-byte-package-save
-                      fop-small-symbol-in-package-save
-                      file)
-           (dump-byte pname-length file))
+           (dump-fop 'fop-keyword-symbol-save file pname-length))
           (t
-           (dump-fop* (dump-package pkg file)
-                      fop-symbol-in-byte-package-save
-                      fop-symbol-in-package-save
-                      file)
-           (dump-word pname-length file)))
+           (dump-fop 'fop-symbol-in-package-save file
+                     (dump-package pkg file) pname-length)))
 
-    #+sb-xc-host (dump-base-chars-of-string pname file)
-    #-sb-xc-host (#!+sb-unicode dump-characters-of-string
-                  #!-sb-unicode dump-base-chars-of-string
-                  pname file)
+    (unless dumped-as-copy
+      #+sb-xc-host (dump-base-chars-of-string pname file)
+      #-sb-xc-host (#!+sb-unicode dump-characters-of-string
+                    #!-sb-unicode dump-base-chars-of-string
+                    pname file)
+      (push s (gethash (symbol-name s) (fasl-output-string=-table file))))
 
     (setf (gethash s (fasl-output-eq-table file))
           (fasl-output-table-free file))
@@ -1090,7 +1105,7 @@
                 (dump-push (cdr entry) fasl-output))
                (:fdefinition
                 (dump-object (cdr entry) fasl-output)
-                (dump-fop 'fop-fdefinition fasl-output))
+                (dump-fop 'fop-fdefn fasl-output))
                (:known-fun
                 (dump-object (cdr entry) fasl-output)
                 (dump-fop 'fop-known-fun fasl-output))))
@@ -1106,20 +1121,12 @@
           (push info-handle (fasl-output-debug-info fasl-output))))
 
       (let ((num-consts (- header-length sb!vm:code-constants-offset)))
-        (cond ((and (< num-consts #x100) (< code-length #x10000))
-               (dump-fop 'fop-small-code fasl-output)
-               (dump-byte num-consts fasl-output)
-               (dump-integer-as-n-bytes code-length (/ sb!vm:n-word-bytes 2) fasl-output))
-              (t
-               (dump-fop 'fop-code fasl-output)
-               (dump-word num-consts fasl-output)
-               (dump-word code-length fasl-output))))
+        (dump-fop 'fop-code fasl-output num-consts code-length))
 
       (dump-segment code-segment code-length fasl-output)
 
       ;; DUMP-FIXUPS does its own internal DUMP-FOPs: the bytes it
-      ;; dumps aren't included in the LENGTH passed to our
-      ;; FOP-CODE/FOP-SMALL-CODE fop.
+      ;; dumps aren't included in the LENGTH passed to FOP-CODE.
       (dump-fixups fixups fasl-output)
 
       #!-(or x86 x86-64)
@@ -1168,7 +1175,7 @@
   (declare (type fasl-output file))
   (dump-push code-handle file)
   (dump-push entry-handle file)
-  (dump-fop* offset fop-byte-alter-code fop-alter-code file)
+  (dump-fop 'fop-alter-code file offset)
   (values))
 
 ;;; Dump the code, constants, etc. for component. We pass in the
@@ -1226,18 +1233,6 @@
   (dump-byte 0 fasl-output)
   (values))
 
-;;; Dump a FOP-FSET to arrange static linkage (at cold init) between
-;;; FUN-NAME and the already-dumped function whose dump handle is
-;;; FUN-DUMP-HANDLE.
-#+sb-xc-host
-(defun fasl-dump-cold-fset (fun-name fun-dump-handle fasl-output)
-  (declare (type fixnum fun-dump-handle))
-  (aver (legal-fun-name-p fun-name))
-  (dump-non-immediate-object fun-name fasl-output)
-  (dump-push fun-dump-handle fasl-output)
-  (dump-fop 'fop-fset fasl-output)
-  (values))
-
 ;;; Compute the correct list of DEBUG-SOURCE structures and backpatch
 ;;; all of the dumped DEBUG-INFO structures. We clear the
 ;;; FASL-OUTPUT-DEBUG-INFO, so that subsequent components with
@@ -1276,40 +1271,56 @@
     (error "attempt to dump invalid structure:~%  ~S~%How did this happen?"
            struct))
   (note-potential-circularity struct file)
-  (aver (%instance-ref struct 0))
   (do* ((length (%instance-length struct))
-        (ntagged (- length (layout-n-untagged-slots (%instance-ref struct 0))))
+        (layout (%instance-layout struct))
+        #!-interleaved-raw-slots
+        (ntagged (- length (layout-n-untagged-slots layout)))
+        #!+interleaved-raw-slots
+        (bitmap (layout-untagged-bitmap layout))
         (circ (fasl-output-circularity-table file))
         ;; last slot first on the stack, so that the layout is on top:
         (index (1- length) (1- index)))
-      ((minusp index)
-       (dump-fop* length fop-small-struct fop-struct file))
-    (let* ((obj (if (>= index ntagged)
+      ((< index sb!vm:instance-data-start)
+       (dump-non-immediate-object layout file)
+       (dump-fop 'fop-struct file length))
+    (let* ((obj #!-interleaved-raw-slots
+                (if (>= index ntagged)
                     (%raw-instance-ref/word struct (- length index 1))
+                    (%instance-ref struct index))
+                #!+interleaved-raw-slots
+                (if (logbitp index bitmap)
+                    (%raw-instance-ref/word struct index)
                     (%instance-ref struct index)))
            (ref (gethash obj circ)))
-      (cond (ref
-             (aver (not (zerop index)))
-             (push (make-circularity :type :struct-set
-                                     :object struct
-                                     :index index
-                                     :value obj
-                                     :enclosing-object ref)
-                   *circularities-detected*)
-             (sub-dump-object nil file))
-            (t
-             (sub-dump-object obj file))))))
+      (sub-dump-object (cond (ref
+                              (push (make-circularity :type :struct-set
+                                                      :object struct
+                                                      :index index
+                                                      :value obj
+                                                      :enclosing-object ref)
+                                    *circularities-detected*)
+                              nil)
+                             (t obj))
+                       file))))
 
 (defun dump-layout (obj file)
   (when (layout-invalid obj)
     (compiler-error "attempt to dump reference to obsolete class: ~S"
                     (layout-classoid obj)))
   (let ((name (classoid-name (layout-classoid obj))))
+    ;; Q: Shouldn't we aver that NAME is the proper name for its classoid?
     (unless name
       (compiler-error "dumping anonymous layout: ~S" obj))
+    ;; The target lisp can save some space in fasls (sometimes),
+    ;; but the cross-compiler can't because we need to construct the
+    ;; cold representation of all layouts, not reference host layouts.
+    #-sb-xc-host
+    (let ((fop (known-layout-fop name)))
+      (when fop
+        (return-from dump-layout (dump-byte fop file))))
     (dump-object name file))
   (sub-dump-object (layout-inherits obj) file)
   (sub-dump-object (layout-depthoid obj) file)
   (sub-dump-object (layout-length obj) file)
-  (sub-dump-object (layout-n-untagged-slots obj) file)
+  (sub-dump-object (layout-raw-slot-metadata obj) file)
   (dump-fop 'fop-layout file))
