@@ -24,6 +24,15 @@
       (make-wired-tn *backend-t-primitive-type*
                      control-stack-arg-scn n)))
 
+(defun standard-arg-location-sc (n)
+  (declare (type unsigned-byte n))
+  (if (< n register-arg-count)
+      (make-sc-offset register-arg-scn
+                      (nth n *register-arg-offsets*))
+      (make-sc-offset control-stack-arg-scn n)))
+
+(defconstant arg-count-sc (make-sc-offset immediate-arg-scn nargs-offset))
+(defconstant closure-sc (make-sc-offset descriptor-reg-sc-number lexenv-offset))
 
 ;;; Make a passing location TN for a local call return PC.  If standard is
 ;;; true, then use the standard (full call) location, otherwise use any legal
@@ -146,9 +155,7 @@
     (storew value frame-pointer (tn-offset variable-home-tn))))
 
 (define-vop (xep-allocate-frame)
-  (:info start-lab copy-more-arg-follows)
-  (:ignore copy-more-arg-follows)
-  (:vop-var vop)
+  (:info start-lab)
   (:temporary (:scs (non-descriptor-reg)) temp)
   (:generator 1
     ;; Make sure the function is aligned, and drop a label pointing to this
@@ -166,8 +173,11 @@
       (inst compute-code-from-lip lip-tn entry-point temp code-tn)
       ;; ### We should also save it on the stack so that the garbage
       ;; collector won't forget about us if we call anyone else.
-      )
-    ;; Build our stack frames.
+      )))
+
+(define-vop (xep-setup-sp)
+  (:vop-var vop)
+  (:generator 1
     (inst addi (* n-word-bytes (sb-allocated-size 'control-stack))
           cfp-tn csp-tn)
     (let ((nfp (current-nfp-tn vop)))
@@ -1037,15 +1047,6 @@ default-value-8
 
 ;;;; XEP hackery:
 
-;;; We don't need to do anything special for regular functions.
-;;;
-(define-vop (setup-environment)
-  (:info label)
-  (:ignore label)
-  (:generator 0
-    ;; Don't bother doing anything.
-    ))
-
 ;;; Get the lexical environment from its passing location.
 ;;;
 (define-vop (setup-closure-environment)
@@ -1202,7 +1203,7 @@ default-value-8
     (inst sub csp-tn count context)))
 
 ;;; Signal wrong argument count error if Nargs isn't = to Count.
-;;;
+#!-precise-arg-count-error
 (define-vop (verify-arg-count)
   (:policy :fast-safe)
   (:translate sb!c::%verify-arg-count)
@@ -1219,6 +1220,28 @@ default-value-8
             (t
              (inst bci :<> nil (fixnumize count) nargs err-lab))))))
 
+#!+precise-arg-count-error
+(define-vop (verify-arg-count)
+  (:policy :fast-safe)
+  (:args (nargs :scs (any-reg)))
+  (:arg-types positive-fixnum (:constant t) (:constant t))
+  (:info min max)
+  (:vop-var vop)
+  (:save-p :compute-only)
+  (:generator 3
+    (let ((err-lab
+           (generate-error-code vop invalid-arg-count-error nargs)))
+      (cond ((not min)
+             (if (zerop max)
+                 (inst bc :<> nil nargs zero-tn err-lab)
+                 (inst bci :<> nil (fixnumize max) nargs err-lab)))
+            (max
+             (when (plusp min)
+               (inst bci :> nil (fixnumize min) nargs err-lab))
+             (inst bci :< nil (fixnumize max) nargs err-lab))
+            ((plusp min)
+             (inst bci :> nil (fixnumize min) nargs err-lab))))))
+
 ;;; Signal argument errors.
 ;;;
 (macrolet ((frob (name error translate &rest args)
