@@ -41,7 +41,6 @@
   ;; legal name?
   (check-fun-name name)
 
-
   ;; KLUDGE: This can happen when eg. compiling a NAMED-LAMBDA, and isn't
   ;; guarded against elsewhere -- so we want to assert package locks here. The
   ;; reason we do it only when stomping on existing stuff is because we want
@@ -74,8 +73,6 @@
   (when (boundp '*free-funs*)       ; when compiling
     (remhash name *free-funs*))
 
-  (note-if-setf-fun-and-macro name)
-
   (values))
 
 ;;; This is called to do something about SETF functions that overlap
@@ -84,34 +81,42 @@
 ;;; warning. Due to the weak semantics of the (SETF FUNCTION) name, we
 ;;; can't assume that they aren't just naming a function (SETF FOO)
 ;;; for the heck of it. NAME is already known to be well-formed.
-(defun note-if-setf-fun-and-macro (name)
-  (when (and (consp name)
-             (eq (car name) 'setf))
-    (when (or (info :setf :inverse (second name))
-              (info :setf :expander (second name)))
+(defun warn-if-setf-macro (name)
+  ;; Never warn about this situation when running the cross-compiler.
+  ;; SBCL provides expanders/inverses *and* functions for most SETFable things
+  ;; even when CLHS does not specifically state that #'(SETF x) exists.
+  #+sb-xc-host (declare (ignore name))
+  #-sb-xc-host
+  (let ((stem (second name)))
+    (when (or (info :setf :inverse stem) (info :setf :expander stem))
       (compiler-style-warn
-       "defining as a SETF function a name that already has a SETF macro:~
-       ~%  ~S"
-       name)))
+         "defining function ~S when ~S already has a SETF macro"
+         name stem)))
   (values))
 
 ;;; Make NAME no longer be a function name: clear everything back to
 ;;; the default.
 (defun undefine-fun-name (name)
   (when name
-    (macrolet ((frob (type &optional val)
-                 `(unless (eq (info :function ,type name) ,val)
-                    (setf (info :function ,type name) ,val))))
-      (frob :info)
-      (frob :type (specifier-type 'function))
-      (frob :where-from :assumed)
-      (frob :inlinep)
-      (frob :kind)
-      (frob :macro-function)
-      (frob :inline-expansion-designator)
-      (frob :source-transform)
-      (frob :structure-accessor)
-      (frob :assumed-type)))
+    (macrolet ((frob (&rest types)
+                 `(clear-info-values
+                   name ',(mapcar (lambda (x)
+                                    (meta-info-number
+                                     (meta-info-or-lose :function x)))
+                                  types))))
+      ;; Note that this does not clear the :DEFINITION.
+      ;; That's correct, because if we lose the association between a
+      ;; symbol and its #<fdefn> object, it could lead to creation of
+      ;; a non-unique #<fdefn> for a name.
+      (frob :info
+            :type
+            :where-from
+            :inlinep
+            :kind
+            :macro-function
+            :inline-expansion-designator
+            :source-transform
+            :assumed-type)))
   (values))
 
 ;;; part of what happens with DEFUN, also with some PCL stuff: Make
@@ -121,16 +126,12 @@
   (when (eq (info :function :where-from name) :assumed)
     (setf (info :function :where-from name) :defined)
     (if (info :function :assumed-type name)
-        (setf (info :function :assumed-type name) nil))))
+        (clear-info :function :assumed-type name))))
 
-;;; Decode any raw (INFO :FUNCTION :INLINE-EXPANSION-DESIGNATOR FUN-NAME)
-;;; value into a lambda expression, or return NIL if there is none.
+;;; Trivially wrap (INFO :FUNCTION :INLINE-EXPANSION-DESIGNATOR FUN-NAME)
 (declaim (ftype (function ((or symbol cons)) list) fun-name-inline-expansion))
 (defun fun-name-inline-expansion (fun-name)
-  (let ((info (info :function :inline-expansion-designator fun-name)))
-    (if (functionp info)
-        (funcall info)
-        info)))
+  (info :function :inline-expansion-designator fun-name))
 
 ;;;; ANSI Common Lisp functions which are defined in terms of the info
 ;;;; database
@@ -162,21 +163,33 @@ only."
   (when (eq (info :function :kind symbol) :special-form)
     (error "~S names a special form." symbol))
   (with-single-package-locked-error (:symbol symbol "setting the macro-function of ~S")
+    (clear-info :function :type symbol)
     (setf (info :function :kind symbol) :macro)
     (setf (info :function :macro-function symbol) function)
-    ;; This is a nice thing to have in the target SBCL, but in the
-    ;; cross-compilation host it's not nice to mess with
-    ;; (SYMBOL-FUNCTION FOO) where FOO might be a symbol in the
-    ;; cross-compilation host's COMMON-LISP package.
-    #-sb-xc-host
-    (setf (symbol-function symbol)
-          (lambda (&rest args)
-            (declare (ignore args))
-            ;; (ANSI specification of FUNCALL says that this should be
-            ;; an error of type UNDEFINED-FUNCTION, not just SIMPLE-ERROR.)
-            (error 'undefined-function :name symbol))))
+    (install-guard-function symbol `(:macro ,symbol) nil))
   function)
 
+;; Set (SYMBOL-FUNCTION SYMBOL) to a closure that signals an error,
+;; preventing funcall/apply of macros and special operators.
+(defun install-guard-function (symbol fun-name docstring)
+  (when docstring
+    (setf (random-documentation symbol 'function) docstring))
+  #-sb-xc-host
+  ;; (SETF SYMBOL-FUNCTION) goes out of its way to disallow this closure,
+  ;; but we can trivially replicate its low-level effect.
+  (setf (fdefn-fun (find-or-create-fdefn symbol))
+        (sb!impl::set-closure-name
+         (lambda (&rest args)
+           (declare (ignore args))
+           ;; ANSI specification of FUNCALL says that this should be
+           ;; an error of type UNDEFINED-FUNCTION, not just SIMPLE-ERROR.
+           ;; SPECIAL-FORM-FUNCTION is a subtype of UNDEFINED-FUNCTION.
+           (error (if (eq (info :function :kind symbol) :special-form)
+                      'special-form-function
+                      'undefined-function)
+                  :name symbol))
+         fun-name)))
+
 (defun fun-locally-defined-p (name env)
   (and env
        (let ((fun (cdr (assoc name (lexenv-funs env) :test #'equal))))
@@ -269,25 +282,42 @@ return NIL. Can be set with SETF when ENV is NIL."
 
 (defun (setf fdocumentation) (string name doc-type)
   (declare (type (or null string) string))
-  (case doc-type
-    (variable (setf (info :variable :documentation name) string))
-    (function
-     ;; KLUDGE: FDEFINITION isn't ready early enough during cold-init, so
-     ;; special case for symbols.
-     (if (symbolp name)
-         (setf (%fun-doc (symbol-function name)) string)
-         (when (legal-fun-name-p name)
-           (setf (%fun-doc (fdefinition name)) string))))
-    (structure (cond
-                 ((eq (info :type :kind name) :instance)
-                  (setf (info :type :documentation name) string))
-                 ((info :typed-structure :info name)
-                  (setf (info :typed-structure :documentation name) string))))
-    (type (setf (info :type :documentation name) string))
-    (setf (setf (info :setf :documentation name) string))
-    (t
-     (when (typep name '(or symbol cons))
-       (setf (random-documentation name doc-type) string))))
+  (let ((info-number
+         (macrolet ((info-number (class type)
+                      (meta-info-number (meta-info-or-lose class type))))
+           (case doc-type
+             (variable (info-number :variable :documentation))
+             (structure
+              (cond ((eq (info :type :kind name) :instance)
+                     (info-number :type :documentation))
+                    ((info :typed-structure :info name)
+                     (info-number :typed-structure :documentation))))
+             (type (info-number :type :documentation))
+             (setf (info-number :setf :documentation))))))
+    (cond (info-number
+           (if string
+               (set-info-value name info-number string)
+               (clear-info-values name (list info-number))))
+          ((eq doc-type 'function)
+           ;; FIXME: this silently loses
+           ;; * (setf (documentation '(a bad name) 'function) "x") => "x"
+           ;; * (documentation '(a bad name) 'function) => NIL
+           ;; which is fine because as noted in pcl/documentation.lsp
+           ;;   even for supported doc types an implementation is permitted
+           ;;   to discard docs at any time
+           ;; but should a warning be issued just as for an unknown DOC-TYPE?
+           ;;
+           ;; And there's additional weirdness if you do, in this order -
+           ;;  * (setf (documentation 'foo 'function) "hi")
+           ;;  * (defun foo () "hey" 1)
+           ;;  * (documentation 'foo 'function) => "hi" ; should be "hey"
+           ;; CLHS says regarding DEFUN:
+           ;; " Documentation is attached as a documentation string to
+           ;;   /name/ (as kind function) and to the /function object/."
+           (when (legal-fun-name-p name)
+             (setf (%fun-doc (fdefinition name)) string)))
+          ((typep name '(or symbol cons))
+           (setf (random-documentation name doc-type) string))))
   string)
 
 (defun random-documentation (name type)
@@ -300,3 +330,19 @@ return NIL. Can be set with SETF when ENV is NIL."
         (push (cons type new-value)
               (info :random-documentation :stuff name))))
   new-value)
+
+;; Return the number of calls to NAME that IR2 emitted as full calls,
+;; not counting calls via #'F that went untracked.
+;; Return 0 if the answer is nonzero but a warning was already signaled
+;; about any full calls were emitted. This return convention satisfies the
+;; intended use of this statistic - to decide whether to generate a warning
+;; about failure to inline NAME, which is shown at most once per name
+;; to avoid unleashing a flood of identical warnings.
+(defun emitted-full-call-count (name)
+  (let ((status (car (info :function :emitted-full-calls name))))
+     (and (integerp status)
+          ;; Bit 0 tells whether any call was NOT in the presence of
+          ;; a 'notinline' declaration, thus eligible to be inline.
+          ;; Bit 1 tells whether any warning was emitted yet.
+          (= (logand status 3) #b01)
+          (ash status -2)))) ; the call count as tracked by IR2
