@@ -153,9 +153,9 @@ extent of the block."
                           :value value-lvar)))
     (when (ctran-deleted-p exit-ctran)
       (throw 'locall-already-let-converted exit-ctran))
-    (push exit (entry-exits entry))
     (setf (lvar-dest value-lvar) exit)
     (ir1-convert start value-ctran value-lvar value)
+    (push exit (entry-exits entry))
     (link-node-to-previous-ctran exit value-ctran)
     (let ((home-lambda (ctran-home-lambda-or-null start)))
       (when home-lambda
@@ -338,6 +338,7 @@ Evaluate the FORMS in the specified SITUATIONS (any of :COMPILE-TOPLEVEL,
                     ,(compile-in-lexenv
                       nil
                       `(lambda (,whole ,environment)
+                         ,@(macro-policy-decls t)
                          ,@local-decls
                          ,body)
                       lexenv))))))))
@@ -515,23 +516,18 @@ Return VALUE without evaluating it."
                      thing))))
 
 (defun fun-name-leaf (thing)
-  (if (consp thing)
-      (cond
-        ((member (car thing)
-                 '(lambda named-lambda lambda-with-lexenv))
-         (values (ir1-convert-lambdalike
-                  thing
-                  :debug-name (name-lambdalike thing))
-                 t))
-        ((legal-fun-name-p thing)
-         (values (find-lexically-apparent-fun
-                  thing "as the argument to FUNCTION")
-                 nil))
-        (t
-         (compiler-error "~S is not a legal function name." thing)))
-      (values (find-lexically-apparent-fun
-               thing "as the argument to FUNCTION")
-              nil)))
+  (cond
+    ((typep thing
+            '(cons (member lambda named-lambda lambda-with-lexenv)))
+     (values (ir1-convert-lambdalike
+              thing :debug-name (name-lambdalike thing))
+             t))
+    ((legal-fun-name-p thing)
+     (values (find-lexically-apparent-fun
+              thing "as the argument to FUNCTION")
+             nil))
+    (t
+     (compiler-error "~S is not a legal function name." thing))))
 
 (def-ir1-translator %%allocate-closures ((&rest leaves) start next result)
   (aver (eq result 'nil))
