@@ -15,6 +15,7 @@
 
 ;;;; LAMBDA hackery
 
+;;;; FIXME: where is that file?
 ;;;; Note: Take a look at the compiler-overview.tex section on "Hairy
 ;;;; function representation" before you seriously mess with this
 ;;;; stuff.
@@ -412,6 +413,9 @@
                 (let ((bindings (default-bindings))
                       (call
                        `(locally
+                         ;; See lengthy comment at top of 'seqtran'
+                         ;; as to why muffling is not done during xc.
+                            #-sb-xc-host
                             (declare (muffle-conditions code-deletion-note))
                           (%funcall ,fun ,@(reverse vals) ,@(default-vals)))))
                   (ir1-convert-lambda-body (if bindings
@@ -434,7 +438,7 @@
     fun))
 
 ;;; This function deals with supplied-p vars in optional arguments. If
-;;; the there is no supplied-p arg, then we just call
+;;; there is no supplied-p arg, then we just call
 ;;; IR1-CONVERT-HAIRY-ARGS on the remaining arguments, and generate a
 ;;; optional entry that calls the result. If there is a supplied-p
 ;;; var, then we add it into the default vars and throw a T into the
@@ -585,13 +589,19 @@
                      (default (arg-info-default info))
                      (keyword (arg-info-key info))
                      (supplied-p (arg-info-supplied-p info))
+                     (supplied-used-p (arg-info-supplied-used-p info))
                      (n-value (sb!xc:gensym "N-VALUE-"))
                      (clause (cond (supplied-p
                                     (let ((n-supplied (sb!xc:gensym "N-SUPPLIED-")))
-                                      (temps n-supplied)
+                                      (temps (list n-supplied
+                                                   (if supplied-used-p
+                                                       nil
+                                                       0)))
                                       (arg-vals n-value n-supplied)
                                       `((eq ,n-key ',keyword)
-                                        (setq ,n-supplied t)
+                                        (setq ,n-supplied ,(if supplied-used-p
+                                                               t
+                                                               1))
                                         (setq ,n-value ,n-value-temp))))
                                    (t
                                     (arg-vals n-value)
@@ -606,13 +616,15 @@
                 (tests clause)))
 
             (unless allowp
-              (temps n-allowp n-lose n-losep)
+              (temps n-allowp
+                     (list n-lose 0)
+                     (list n-losep 0))
               (unless found-allow-p
                 (tests `((eq ,n-key :allow-other-keys)
                          (setq ,n-allowp ,n-value-temp))))
               (tests `(t
                        (setq ,n-lose ,n-key
-                             ,n-losep t))))
+                             ,n-losep 1))))
 
             (body
              `(when (oddp ,n-count)
@@ -640,7 +652,7 @@
                     (cond ,@(tests))))))
 
             (unless allowp
-              (body `(when (and ,n-losep (not ,n-allowp))
+              (body `(when (and (/= ,n-losep 0) (not ,n-allowp))
                        (%unknown-key-arg-error ,n-lose)))))))
 
       (let ((ep (ir1-convert-lambda-body
@@ -727,11 +739,19 @@
                  (unless supplied-p
                    (setf (arg-info-supplied-p info) supplied-temp))
                  (when hairy-default
-                   (setf (arg-info-default info) nil))
+                   (setf (arg-info-default info) nil)
+                   (unless supplied-p
+                     (setf (arg-info-supplied-used-p info) nil)))
                  (main-vars supplied-temp)
                  (cond (hairy-default
-                        (main-vals nil nil)
-                        (bind-vals `(if ,n-supplied ,n-val ,default)))
+                        (main-vals nil
+                                   (if supplied-p
+                                       nil
+                                       0))
+                        (bind-vals
+                         (if supplied-p
+                             `(if ,n-supplied ,n-val ,default)
+                             `(if (eq ,n-supplied 0) ,default ,n-val))))
                        (t
                         (main-vals default nil)
                         (bind-vals n-val)))
@@ -1043,10 +1063,23 @@
            (let ((defined-fun-res (get-defined-fun name (second lambda-expression)))
                  (res (ir1-convert-lambda lambda-expression
                                           :maybe-add-debug-catch t
-                                          :source-name name)))
+                                          :source-name name))
+                 (info (info :function :info name)))
              (assert-global-function-definition-type name res)
              (push res (defined-fun-functionals defined-fun-res))
-             (unless (eq (defined-fun-inlinep defined-fun-res) :notinline)
+             (unless (or
+                      (eq (defined-fun-inlinep defined-fun-res) :notinline)
+                      ;; Don't treat recursive stubs like CAR as self-calls
+                      ;; Maybe just use the fact that it is a known function?
+                      ;; Though a known function may be used
+                      ;; because of some other attributues but
+                      ;; still wants to get optimized self calls
+                      (and info
+                           (or (fun-info-templates info)
+                               (fun-info-transforms info)
+                               (fun-info-ltn-annotate info)
+                               (fun-info-ir2-convert info)
+                               (fun-info-optimizer info))))
                (substitute-leaf-if
                 (lambda (ref)
                   (policy ref (> recognize-self-calls 0)))
@@ -1168,12 +1201,12 @@
     (assert-definition-type
      fun type
      ;; KLUDGE: Common Lisp is such a dynamic language that in general
-     ;; all we can do here in general is issue a STYLE-WARNING. It
-     ;; would be nice to issue a full WARNING in the special case of
-     ;; of type mismatches within a compilation unit (as in section
-     ;; 3.2.2.3 of the spec) but at least as of sbcl-0.6.11, we don't
-     ;; keep track of whether the mismatched data came from the same
-     ;; compilation unit, so we can't do that. -- WHN 2001-02-11
+     ;; all we can do here is issue a STYLE-WARNING. It would be nice
+     ;; to issue a full WARNING in the special case of type mismatches
+     ;; within a compilation unit (as in section 3.2.2.3 of the spec)
+     ;; but at least as of sbcl-0.6.11, we don't keep track of whether
+     ;; the mismatched data came from the same compilation unit, so we
+     ;; can't do that. -- WHN 2001-02-11
      :lossage-fun #'compiler-style-warn
      :unwinnage-fun (cond (info #'compiler-style-warn)
                           (for-real #'compiler-notify)
@@ -1278,6 +1311,30 @@
 
   (values))
 
+;; Similar to above, detect duplicate definitions within a file,
+;; but the package lock check is unnecessary - it's handled elsewhere.
+;;
+;; Additionally, this is a STYLE-WARNING, not a WARNING, because there is
+;; meaningful behavior that can be ascribed to some redefinitions, e.g.
+;;  (defmacro foo () first-definition)
+;;  (defun f () (use-it (foo )))
+;;  (defmacro foo () other-definition)
+;; will use the first definition when compiling F, but make the second available
+;; in the loaded fasl. In this usage it would have made sense to wrap the
+;; respective definitions with EVAL-WHEN for different situations,
+;; but as long as the compile-time behavior is deterministic, it's just bad style
+;; and not flat-out wrong, though there is indeed some waste in the fasl.
+;;
+;; KIND is the globaldb KIND of this NAME
+(defun %compiler-defmacro (kind name compile-toplevel)
+  (when compile-toplevel
+    (let ((name-key `(,kind ,name)))
+      (when (boundp '*lexenv*)
+        (aver (fasl-output-p *compile-object*))
+        (if (member name-key *fun-names-in-this-file* :test #'equal)
+            (compiler-style-warn 'same-file-redefinition-warning :name name)
+            (push name-key *fun-names-in-this-file*))))))
+
 
 ;;; Entry point utilities
 
