@@ -270,7 +270,7 @@
       (let ((kind (info :variable :kind name))
             (type (info :variable :type name))
             (where-from (info :variable :where-from name)))
-        (when (eq kind :unknown)
+        (when (and (eq kind :unknown) (not (check-deprecated-variable name)))
           (note-undefined-reference name :variable))
         (setf (gethash name *free-vars*)
               (case kind
@@ -357,17 +357,24 @@
                     ;; functions don't work on general instances, only on
                     ;; STRUCTURE!OBJECTs.
                     ;;
+                    ;; Behold the wonderfully clear sense of this-
+                    ;;  WHEN (EMIT-MAKE-LOAD-FORM VALUE)
+                    ;; meaning "when you're _NOT_ using a custom load-form"
+                    ;;
                     ;; FIXME: What about funcallable instances with
                     ;; user-defined MAKE-LOAD-FORM methods?
                     (when (emit-make-load-form value)
-                      (dotimes (i (- (%instance-length value)
-                                     #+sb-xc-host 0
-                                     #-sb-xc-host (layout-n-untagged-slots
-                                                   (%instance-ref value 0))))
+                      #+sb-xc-host
+                      (aver (zerop (layout-raw-slot-metadata
+                                    (%instance-layout value))))
+                      (do-instance-tagged-slot (i value)
                         (grovel (%instance-ref value i)))))
+                   ;; The cross-compiler can dump certain instances that are not
+                   ;; subtypes of STRUCTURE!OBJECT, as long as it has processed
+                   ;; the defstruct.
                    #+sb-xc-host
                    ((satisfies sb!kernel::xc-dumpable-structure-instance-p)
-                    (dotimes (i (%instance-length value))
+                    (do-instance-tagged-slot (i value)
                       (grovel (%instance-ref value i))))
                    (t
                     (compiler-error
@@ -542,7 +549,7 @@
 
 ;;;; IR1-CONVERT, macroexpansion and special form dispatching
 
-(declaim (ftype (sfunction (ctran ctran (or lvar null) t &optional t)
+(declaim (ftype (sfunction (ctran ctran (or lvar null) t)
                            (values))
                 ir1-convert))
 (macrolet (;; Bind *COMPILER-ERROR-BAILOUT* to a function that throws
@@ -572,9 +579,9 @@
   ;; the creation using backquote of forms that contain leaf
   ;; references, without having to introduce dummy names into the
   ;; namespace.
-  (defun ir1-convert (start next result form &optional alias)
+  (defun ir1-convert (start next result form)
     (ir1-error-bailout (start next result form)
-      (let* ((*current-path* (ensure-source-path (or alias form)))
+      (let* ((*current-path* (ensure-source-path form))
              (start (instrument-coverage start nil form)))
         (cond ((atom form)
                (cond ((and (symbolp form) (not (keywordp form)))
@@ -682,24 +689,30 @@
                          `(symbol-value ',name)))
         (etypecase var
           (leaf
-           (when (lambda-var-p var)
-             (let ((home (ctran-home-lambda-or-null start)))
-               (when home
-                 (sset-adjoin var (lambda-calls-or-closes home))))
-             (when (lambda-var-ignorep var)
-               ;; (ANSI's specification for the IGNORE declaration requires
-               ;; that this be a STYLE-WARNING, not a full WARNING.)
-               #-sb-xc-host
-               (compiler-style-warn "reading an ignored variable: ~S" name)
-               ;; there's no need for us to accept ANSI's lameness when
-               ;; processing our own code, though.
-               #+sb-xc-host
-               (warn "reading an ignored variable: ~S" name)))
-           (when (global-var-p var)
-             (check-deprecated-variable name))
+           (cond
+             ((lambda-var-p var)
+              (let ((home (ctran-home-lambda-or-null start)))
+                (when home
+                  (sset-adjoin var (lambda-calls-or-closes home))))
+              (when (lambda-var-ignorep var)
+                ;; (ANSI's specification for the IGNORE declaration requires
+                ;; that this be a STYLE-WARNING, not a full WARNING.)
+                #-sb-xc-host
+                (compiler-style-warn "reading an ignored variable: ~S" name)
+                ;; there's no need for us to accept ANSI's lameness when
+                ;; processing our own code, though.
+                #+sb-xc-host
+                (warn "reading an ignored variable: ~S" name)))
+             (t
+              (multiple-value-bind (state since replacements)
+                  (check-deprecated-variable name)
+                (when (eq state :final)
+                  (ir1-convert
+                   start next result
+                   `(deprecation-error ,since ',name '(,@replacements)))
+                  (return-from ir1-convert-var (values))))))
            (reference-leaf start next result var name))
-          (cons
-           (aver (eq (car var) 'macro))
+          ((cons (eql macro)) ; symbol-macro
            ;; FIXME: [Free] type declarations. -- APD, 2002-01-26
            (ir1-convert start next result (cdr var)))
           (heap-alien-info
@@ -1082,7 +1095,19 @@
         (let* ((name (leaf-source-name var))
                (transform (info :function :source-transform name)))
           (if transform
-              (multiple-value-bind (transformed pass) (funcall transform form)
+              (multiple-value-bind (transformed pass)
+                  (if (functionp transform)
+                      (funcall transform form)
+                      (let ((result
+                             (if (eq (cdr transform) :predicate)
+                                 (and (singleton-p (cdr form))
+                                      `(%instance-typep
+                                        ,(cadr form)
+                                        ',(dd-name (car transform))))
+                                 (slot-access-transform
+                                  (if (consp name) :write :read)
+                                  (cdr form) transform))))
+                        (values result (null result))))
                 (cond (pass
                        (ir1-convert-maybe-predicate start next result form var))
                       (t
@@ -1385,7 +1410,7 @@
 
 ;;; Process an ignore/ignorable declaration, checking for various losing
 ;;; conditions.
-(defun process-ignore-decl (spec vars fvars)
+(defun process-ignore-decl (spec vars fvars lexenv)
   (declare (list spec vars fvars))
   (dolist (name (rest spec))
     (let ((var (find-in-bindings-or-fbindings name vars fvars)))
@@ -1402,12 +1427,18 @@
                   (:special "a special variable")
                   (:global "a global lexical variable")
                   (:alien "a global alien variable")
-                  (t "an unknown variable"))
+                  (t (if (assoc name (lexenv-vars lexenv))
+                         "a variable from outer scope"
+                         "an unknown variable")))
                 name)
                (values
-                (if (info :function :kind (second name))
-                    "a global function"
-                    "an unknown function")
+                (cond ((assoc (second name) (lexenv-funs lexenv)
+                              :test #'equal)
+                       "a function from outer scope")
+                      ((info :function :kind (second name))
+                       "a global function")
+                      (t
+                       "an unknown function"))
                 (second name)))))
         ((and (consp var) (eq (car var) 'macro))
          ;; Just ignore the IGNORE decl: we don't currently signal style-warnings
@@ -1504,6 +1535,7 @@
   (declare (type list raw-spec vars fvars))
   (declare (type lexenv res))
   (let ((spec (canonized-decl-spec raw-spec))
+        (optimize-qualities)
         (result-type *wild-type*))
     (values
      (case (first spec)
@@ -1515,12 +1547,13 @@
        ((inline notinline maybe-inline)
         (process-inline-decl spec res fvars))
        ((ignore ignorable)
-        (process-ignore-decl spec vars fvars)
+        (process-ignore-decl spec vars fvars res)
         res)
        (optimize
-        (make-lexenv
-         :default res
-         :policy (process-optimize-decl spec (lexenv-policy res))))
+        (multiple-value-bind (new-policy specified-qualities)
+            (process-optimize-decl spec (lexenv-policy res))
+          (setq optimize-qualities specified-qualities)
+          (make-lexenv :default res :policy new-policy)))
        (muffle-conditions
         (make-lexenv
          :default res
@@ -1557,7 +1590,8 @@
           (if fn
               (funcall fn res spec vars fvars)
               res))))
-     result-type)))
+     result-type
+     optimize-qualities)))
 
 ;;; Use a list of DECLARE forms to annotate the lists of LAMBDA-VAR
 ;;; and FUNCTIONAL structures which are being bound. In addition to
@@ -1574,23 +1608,31 @@
                       (lexenv *lexenv*) (binding-form-p nil) (context :compile))
   (declare (list decls vars fvars))
   (let ((result-type *wild-type*)
+        (optimize-qualities)
         (*post-binding-variable-lexenv* nil))
     (dolist (decl decls)
       (dolist (spec (rest decl))
-        (progv
+        (flet
+            ((process-it ()
+               (unless (consp spec)
+                 (compiler-error "malformed declaration specifier ~S in ~S"
+                                 spec decl))
+               (multiple-value-bind (new-env new-result-type new-qualities)
+                   (process-1-decl spec lexenv vars fvars binding-form-p context)
+                 (setq lexenv new-env
+                       optimize-qualities (nconc new-qualities optimize-qualities))
+                 (unless (eq new-result-type *wild-type*)
+                   (setq result-type
+                         (values-type-intersection result-type new-result-type))))))
+          (if (eq context :compile)
+              (let ((*current-path* (or (get-source-path spec)
+                                        (get-source-path decl)
+                                        *current-path*)))
+                (process-it))
             ;; Kludge: EVAL calls this function to deal with LOCALLY.
-            (when (eq context :compile) (list '*current-path*))
-            (when (eq context :compile) (list (or (get-source-path spec)
-                                                  (get-source-path decl)
-                                                  *current-path*)))
-          (unless (consp spec)
-            (compiler-error "malformed declaration specifier ~S in ~S" spec decl))
-          (multiple-value-bind (new-env new-result-type)
-              (process-1-decl spec lexenv vars fvars binding-form-p context)
-            (setq lexenv new-env)
-            (unless (eq new-result-type *wild-type*)
-              (setq result-type
-                    (values-type-intersection result-type new-result-type)))))))
+              (process-it)))))
+    (advise-if-repeated-optimize-qualities
+     (lexenv-policy lexenv) optimize-qualities)
     (values lexenv result-type *post-binding-variable-lexenv*)))
 
 (defun %processing-decls (decls vars fvars ctran lvar binding-form-p fun)
