@@ -22,8 +22,15 @@
   (do ((lexenv (node-lexenv node)
                (lambda-call-lexenv (lexenv-lambda lexenv))))
       ((null lexenv) nil)
-    (let ((cup (lexenv-cleanup lexenv)))
-      (when cup (return cup)))))
+    (awhen (lexenv-cleanup lexenv)
+      (return it))))
+
+(defun map-nested-cleanups (function block &optional return-value)
+  (declare (type cblock block))
+  (do ((cleanup (block-end-cleanup block)
+                (node-enclosing-cleanup (cleanup-mess-up cleanup))))
+      ((not cleanup) return-value)
+    (funcall function cleanup)))
 
 ;;; Convert the FORM in a block inserted between BLOCK1 and BLOCK2 as
 ;;; an implicit MV-PROG1. The inserted block is returned. NODE is used
@@ -1063,24 +1070,22 @@
 ;;; have the same cleanup info, corresponding to the start, so the
 ;;; same approach returns safe result.
 (defun map-block-nlxes (fun block &optional dx-cleanup-fun)
-  (loop for cleanup = (block-end-cleanup block)
-        then (node-enclosing-cleanup (cleanup-mess-up cleanup))
-        while cleanup
-        do (let ((mess-up (cleanup-mess-up cleanup)))
-             (case (cleanup-kind cleanup)
-               ((:block :tagbody)
-                (aver (entry-p mess-up))
-                (loop for exit in (entry-exits mess-up)
-                      for nlx-info = (exit-nlx-info exit)
-                      do (funcall fun nlx-info)))
-               ((:catch :unwind-protect)
-                (aver (combination-p mess-up))
-                (let* ((arg-lvar (first (basic-combination-args mess-up)))
-                       (nlx-info (constant-value (ref-leaf (lvar-use arg-lvar)))))
-                (funcall fun nlx-info)))
-               ((:dynamic-extent)
-                (when dx-cleanup-fun
-                  (funcall dx-cleanup-fun cleanup)))))))
+  (do-nested-cleanups (cleanup block)
+    (let ((mess-up (cleanup-mess-up cleanup)))
+      (case (cleanup-kind cleanup)
+        ((:block :tagbody)
+         (aver (entry-p mess-up))
+         (loop for exit in (entry-exits mess-up)
+            for nlx-info = (exit-nlx-info exit)
+            do (funcall fun nlx-info)))
+        ((:catch :unwind-protect)
+         (aver (combination-p mess-up))
+         (let* ((arg-lvar (first (basic-combination-args mess-up)))
+                (nlx-info (constant-value (ref-leaf (lvar-use arg-lvar)))))
+           (funcall fun nlx-info)))
+        ((:dynamic-extent)
+         (when dx-cleanup-fun
+           (funcall dx-cleanup-fun cleanup)))))))
 
 ;;; Set the FLAG for all the blocks in COMPONENT to NIL, except for
 ;;; the head and tail which are set to T.
@@ -1577,24 +1582,28 @@
 
 ;;; If any of the VARS in FUN was never referenced and was not
 ;;; declared IGNORE, then complain.
-(defun note-unreferenced-vars (fun)
-  (declare (type clambda fun))
-  (dolist (var (lambda-vars fun))
+(defun note-unreferenced-vars (vars policy)
+  (dolist (var vars)
     (unless (or (leaf-ever-used var)
                 (lambda-var-ignorep var))
-      (let ((*compiler-error-context* (lambda-bind fun)))
-        (unless (policy *compiler-error-context* (= inhibit-warnings 3))
-          ;; ANSI section "3.2.5 Exceptional Situations in the Compiler"
-          ;; requires this to be no more than a STYLE-WARNING.
-          #-sb-xc-host
-          (compiler-style-warn "The variable ~S is defined but never used."
-                               (leaf-debug-name var))
-          ;; There's no reason to accept this kind of equivocation
-          ;; when compiling our own code, though.
-          #+sb-xc-host
-          (warn "The variable ~S is defined but never used."
-                (leaf-debug-name var)))
-        (setf (leaf-ever-used var) t)))) ; to avoid repeated warnings? -- WHN
+      (unless (policy policy (= inhibit-warnings 3))
+        ;; ANSI section "3.2.5 Exceptional Situations in the Compiler"
+        ;; requires this to be no more than a STYLE-WARNING.
+        #-sb-xc-host
+        (compiler-style-warn "The variable ~S is defined but never used."
+                             (leaf-debug-name var))
+        ;; There's no reason to accept this kind of equivocation
+        ;; when compiling our own code, though.
+        #+sb-xc-host
+        (warn "The variable ~S is defined but never used."
+              (leaf-debug-name var)))
+      (setf (leaf-ever-used var) t)))) ; to avoid repeated warnings? -- WHN
+
+(defun note-unreferenced-fun-vars (fun)
+  (declare (type clambda fun))
+  (let ((*compiler-error-context* (lambda-bind fun)))
+    (note-unreferenced-vars (lambda-vars fun)
+                            *compiler-error-context*))
   (values))
 
 (defvar *deletion-ignored-objects* '(t nil))
@@ -2094,10 +2103,15 @@ is :ANY, the function name is not checked."
 ;;; Return the source name of a combination -- or signals an error
 ;;; if the function leaf is anonymous.
 (defun combination-fun-source-name (combination &optional (errorp t))
-  (let ((leaf (ref-leaf (lvar-uses (combination-fun combination)))))
-    (if (or errorp (leaf-has-source-name-p leaf))
-        (values (leaf-source-name leaf) t)
-        (values nil nil))))
+  (let ((uses (principal-lvar-use (combination-fun combination)))
+        leaf)
+    (cond ((and (ref-p uses)
+                (leaf-has-source-name-p (setf leaf (ref-leaf uses))))
+           (values (leaf-source-name leaf) t))
+          (errorp
+           (aver (not "COMBINATION-FUN is not a ref to a nameful leaf")))
+          (t
+           (values nil nil)))))
 
 (defun combination-fun-debug-name (combination)
   (leaf-debug-name (ref-leaf (lvar-uses (combination-fun combination)))))
