@@ -27,11 +27,12 @@
                               &body body)
   (let ((fn-name (symbolicate "IR1-CONVERT-" name)))
     (with-unique-names (whole-var n-env)
-      (multiple-value-bind (body decls #-sb-xc-host doc)
+      (multiple-value-bind (body decls doc)
           (parse-defmacro lambda-list whole-var body name "special form"
                           :environment n-env
                           :error-fun 'compiler-error
                           :wrap-block nil)
+        (declare (ignorable doc))
         `(progn
            (declaim (ftype (function (ctran ctran (or lvar null) t) (values))
                            ,fn-name))
@@ -41,26 +42,16 @@
              ,@decls
              ,body
              (values))
+           (install-guard-function ',name '(:special ,name) ,(or #!+sb-doc doc))
+           ;; The guard function is a closure, which can't have its lambda-list
+           ;; changed independently of other closures based on the same
+           ;; simple-fun. So change it in the ir1-translator since the actual
+           ;; lambda-list is not terribly useful.
            #-sb-xc-host
-           ;; It's nice to do this for error checking in the target
-           ;; SBCL, but it's not nice to do this when we're running in
-           ;; the cross-compilation host Lisp, which owns the
-           ;; SYMBOL-FUNCTION of its COMMON-LISP symbols. These guard
-           ;; functions also provide the documentation for special forms.
-           ;; FIXME: should be disallowed after bootstrap. Package-lock
-           ;; prevents it, but the protection could be stronger than that.
-           ;; The lambda name has significance to COERCE-SYMBOL-TO-FUN.
-           (let ((fun (named-lambda (sb!impl::special-operator ,name)
-                          (&rest args)
-                        ,@(when doc (list doc))
-                        (declare (ignore args))
-                        (error 'special-form-function :name ',name)) ))
-             ;; Set up a macro lambda list to a function.
-             (setf (%simple-fun-arglist fun)
-                   ',(if (eq (first lambda-list) '&whole)
-                         (cddr lambda-list)
-                         lambda-list)
-                   (symbol-function ',name) fun))
+           (setf (%fun-lambda-list #',fn-name) ; This is for DESCRIBE et. al.
+                 ',(if (eq (first lambda-list) '&whole)
+                       (cddr lambda-list)
+                       lambda-list))
            ;; FIXME: Evidently "there can only be one!" -- we overwrite any
            ;; other :IR1-CONVERT value. This deserves a warning, I think.
            (setf (info :function :ir1-convert ',name) #',fn-name)
@@ -88,7 +79,7 @@
 ;;; If the desirability of the transformation depends on the current
 ;;; OPTIMIZE parameters, then the POLICY macro should be used to
 ;;; determine when to pass.
-(defmacro source-transform-lambda (lambda-list &body body)
+(defmacro define-source-transform (fun-name lambda-list &body body)
   (with-unique-names (whole-var n-env name)
     (multiple-value-bind (body decls)
         (parse-defmacro lambda-list whole-var body "source transform" "form"
@@ -98,13 +89,11 @@
                                       (return-from ,name
                                         (values nil t)))
                         :wrap-block nil)
-      `(lambda (,whole-var &aux (,n-env *lexenv*))
-         ,@decls
-         (block ,name
-           ,body)))))
-(defmacro define-source-transform (name lambda-list &body body)
-  `(setf (info :function :source-transform ',name)
-         (source-transform-lambda ,lambda-list ,@body)))
+      `(setf (info :function :source-transform ',fun-name)
+             (lambda (,whole-var &aux (,n-env *lexenv*))
+               ,@decls
+               (block ,name
+                 ,body))))))
 
 ;;;; boolean attribute utilities
 ;;;;
@@ -325,9 +314,11 @@
                                   `(,(car k)
                                      (find-keyword-lvar ,tail ',(cdr k))))
                                 keys))
-                (append (nset-difference (mapcar #'car (binds)) all-dummies)
-                        (mapcar #'car keys))
-                (intersection (mapcar #'car (binds)) (cdr all-dummies)))))))
+                (sort (append (nset-difference (mapcar #'car (binds)) all-dummies)
+                               (mapcar #'car keys))
+                      #'string<)
+                (sort (intersection (mapcar #'car (binds)) (cdr all-dummies))
+                      #'string<))))))
 ) ; EVAL-WHEN
 
 ;;;; DEFTRANSFORM
@@ -477,6 +468,7 @@
                     (list name))
               '(sfunction ,arg-types ,result-type)
               (ir1-attributes ,@attributes)
+              (source-location)
               ,@keys))
 
 ;;; Create a function which parses combination args according to WHAT
@@ -678,6 +670,12 @@
            while ,node-var
            do (progn ,@body))))
 
+(defmacro do-nested-cleanups ((cleanup-var block &optional return-value)
+                              &body body)
+  `(block nil
+     (map-nested-cleanups
+      (lambda (,cleanup-var) ,@body) ,block ,return-value)))
+
 ;;; Bind the IR1 context variables to the values associated with NODE,
 ;;; so that new, extra IR1 conversion related to NODE can be done
 ;;; after the original conversion pass has finished.
