@@ -547,6 +547,7 @@ necessary, since type inference may take arbitrarily long to converge.")
     (when (or (ir2-component-values-receivers (component-info component))
               (component-dx-lvars component))
       (maybe-mumble "stack ")
+      (find-dominators component)
       (stack-analyze component)
       ;; Assign BLOCK-NUMBER for any cleanup blocks introduced by
       ;; stack analysis. There shouldn't be any unreachable code after
@@ -690,6 +691,11 @@ necessary, since type inference may take arbitrarily long to converge.")
         (print-loop (component-outer-loop component))))
     |#
 
+    ;; This should happen at some point before PHYSENV-ANALYZE, and
+    ;; after RECORD-COMPONENT-XREFS.  Beyond that, I haven't really
+    ;; thought things through.  -- AJB, 2014-Jun-08
+    (eliminate-dead-code component)
+
     ;; FIXME: What is MAYBE-MUMBLE for? Do we need it any more?
     (maybe-mumble "env ")
     (physenv-analyze component)
@@ -814,14 +820,18 @@ necessary, since type inference may take arbitrarily long to converge.")
   ;; the file positions that reading of each form started at (i.e. the
   ;; end of the previous form)
   (forms (make-array 10 :fill-pointer 0 :adjustable t) :type (vector t))
-  (positions (make-array 10 :fill-pointer 0 :adjustable t) :type (vector t)))
+  (positions (make-array 10 :fill-pointer 0 :adjustable t) :type (vector t))
+  ;; A vector of character ranges than span each subform in the TLF,
+  ;; reset to empty for each one, updated by form-tracking-stream-observer.
+  (subforms nil :type (or null (vector t)) :read-only t))
 
 ;;; The SOURCE-INFO structure provides a handle on all the source
 ;;; information for an entire compilation.
 (def!struct (source-info
              #-no-ansi-print-object
              (:print-object (lambda (s stream)
-                              (print-unreadable-object (s stream :type t))))
+                              (print-unreadable-object
+                                  (s stream :type t :identity t))))
              (:copier nil))
   ;; the UT that compilation started at
   (start-time (get-universal-time) :type unsigned-byte)
@@ -832,17 +842,23 @@ necessary, since type inference may take arbitrarily long to converge.")
   ;; the stream that we are using to read the FILE-INFO, or NIL if
   ;; no stream has been opened yet
   (stream nil :type (or stream null))
+  ;; for coalescing DEFINITION-SOURCE-LOCATION of effectively toplevel forms
+  ;; inside one truly toplevel form.
+  (last-defn-source-loc)
   ;; if the current compilation is recursive (e.g., due to EVAL-WHEN
   ;; processing at compile-time), the invoking compilation's
   ;; source-info.
   (parent nil :type (or source-info null)))
 
 ;;; Given a pathname, return a SOURCE-INFO structure.
-(defun make-file-source-info (file external-format)
+(defun make-file-source-info (file external-format &optional form-tracking-p)
   (make-source-info
    :file-info (make-file-info :name (truename file)
                               :untruename (merge-pathnames file)
                               :external-format external-format
+                              :subforms
+                              (if form-tracking-p
+                                  (make-array 100 :fill-pointer 0 :adjustable t))
                               :write-date (file-write-date file))))
 
 ;;; Return a SOURCE-INFO to describe the incremental compilation of FORM.
@@ -865,39 +881,6 @@ necessary, since type inference may take arbitrarily long to converge.")
                 (pathnamep (file-info-name finfo)))
             finfo))))
 
-;;; Return a form read from STREAM; or for EOF use the trick,
-;;; popularized by Kent Pitman, of returning STREAM itself. If an
-;;; error happens, then convert it to standard abort-the-compilation
-;;; error condition (possibly recording some extra location
-;;; information).  CONDITION-NAME is what to signal on error,
-;;; and should be INPUT-ERROR-IN-COMPILE-FILE or a subclass of it.
-;;; The signaled condition encapsulates a reader condition.
-(defun read-for-compile-file (stream position condition-name)
-  (handler-case
-      (read-preserving-whitespace stream nil stream)
-    (reader-error (condition)
-      (compiler-error condition-name
-                      ;; We don't need to supply :POSITION here because
-                      ;; READER-ERRORs already know their position in the file.
-                      :condition condition
-                      :stream stream))
-    ;; ANSI, in its wisdom, says that READ should return END-OF-FILE
-    ;; (and that this is not a READER-ERROR) when it encounters end of
-    ;; file in the middle of something it's trying to read.
-    (end-of-file (condition)
-      (compiler-error condition-name
-                      :condition condition
-                      ;; We need to supply :POSITION here because the END-OF-FILE
-                      ;; condition doesn't carry the position that the user
-                      ;; probably cares about, where the failed READ began.
-                      :position position
-                      :stream stream))
-    (error (condition)
-      (compiler-error condition-name
-                      :condition condition
-                      :position position
-                      :stream stream))))
-
 ;;; If STREAM is present, return it, otherwise open a stream to the
 ;;; current file. There must be a current file.
 ;;;
@@ -918,8 +901,17 @@ necessary, since type inference may take arbitrarily long to converge.")
         (setf sb!xc:*compile-file-truename* name
               sb!xc:*compile-file-pathname* (file-info-untruename file-info)
               (source-info-stream info)
-              (open name :direction :input
-                    :external-format external-format)))))
+              (let ((stream
+                     (open name
+                           :direction :input
+                           :external-format external-format
+                           ;; SBCL stream classes aren't available in the host
+                           #-sb-xc-host :class
+                           #-sb-xc-host 'form-tracking-stream)))
+                (when (file-info-subforms file-info)
+                  (setf (form-tracking-stream-observer stream)
+                        (make-form-tracking-stream-observer file-info)))
+                stream)))))
 
 ;;; Close the stream in INFO if it is open.
 (defun close-source-info (info)
@@ -929,6 +921,65 @@ necessary, since type inference may take arbitrarily long to converge.")
   (setf (source-info-stream info) nil)
   (values))
 
+;; Loop over forms read from INFO's stream, calling FUNCTION with each.
+;; CONDITION-NAME is signaled if there is a reader error, and should be
+;; a subtype of not-so-aptly-named INPUT-ERROR-IN-COMPILE-FILE.
+(defun %do-forms-from-info (function info condition-name)
+  (declare (function function))
+  (let* ((file-info (source-info-file-info info))
+         (stream (get-source-stream info))
+         (pos (file-position stream))
+         (form
+          ;; Return a form read from STREAM; or for EOF use the trick,
+          ;; popularized by Kent Pitman, of returning STREAM itself.
+          (handler-case
+              (progn
+                ;; Reset for a new toplevel form.
+                (when (form-tracking-stream-p stream)
+                  (setf (form-tracking-stream-form-start-char-pos stream) nil))
+                (awhen (file-info-subforms file-info)
+                  (setf (fill-pointer it) 0))
+                (read-preserving-whitespace stream nil stream))
+            (reader-error (condition)
+              (compiler-error condition-name
+                ;; We don't need to supply :POSITION here because
+                ;; READER-ERRORs already know their position in the file.
+                              :condition condition
+                              :stream stream))
+            ;; ANSI, in its wisdom, says that READ should return END-OF-FILE
+            ;; (and that this is not a READER-ERROR) when it encounters end of
+            ;; file in the middle of something it's trying to read,
+            ;; making it unfortunately indistinguishable from legal EOF.
+            ;; Were it not for that, it would be more elegant to just
+            ;; handle one more condition in the HANDLER-CASE.
+            ((or end-of-file error) (condition)
+              (compiler-error
+               condition-name
+               :condition condition
+                ;; We need to supply :POSITION here because the END-OF-FILE
+                ;; condition doesn't carry the position that the user
+                ;; probably cares about, where the failed READ began.
+               :position
+               (or (and (form-tracking-stream-p stream)
+                        (form-tracking-stream-form-start-byte-pos stream))
+                   pos)
+               :line/col
+               (and (form-tracking-stream-p stream)
+                    (line/col-from-charpos
+                     stream
+                     (form-tracking-stream-form-start-char-pos stream)))
+               :stream stream)))))
+    (unless (eq form stream) ; not EOF
+      (funcall function form
+               :current-index
+               (let* ((forms (file-info-forms file-info))
+                      (current-idx (+ (fill-pointer forms)
+                                      (file-info-source-root file-info))))
+                 (vector-push-extend form forms)
+                 (vector-push-extend pos (file-info-positions file-info))
+                 current-idx))
+      (%do-forms-from-info function info condition-name))))
+
 ;;; Loop over FORMS retrieved from INFO.  Used by COMPILE-FILE and
 ;;; LOAD when loading from a FILE-STREAM associated with a source
 ;;; file.  ON-ERROR is the name of a condition class that should
@@ -939,22 +990,9 @@ necessary, since type inference may take arbitrarily long to converge.")
   (aver (symbolp form))
   (once-only ((info info))
     `(let ((*source-info* ,info))
-       (loop (destructuring-bind (,form &key ,@keys &allow-other-keys)
-                 (let* ((file-info (source-info-file-info ,info))
-                        (stream (get-source-stream ,info))
-                        (pos (file-position stream))
-                        (form (read-for-compile-file stream pos ,on-error)))
-                   (if (eq form stream) ; i.e., if EOF
-                       (return)
-                       (let* ((forms (file-info-forms file-info))
-                              (current-idx (+ (fill-pointer forms)
-                                              (file-info-source-root
-                                               file-info))))
-                         (vector-push-extend form forms)
-                         (vector-push-extend pos (file-info-positions
-                                                  file-info))
-                         (list form :current-index current-idx))))
-               ,@body)))))
+       (%do-forms-from-info (lambda (,form &key ,@keys &allow-other-keys)
+                              ,@body)
+                            ,info ,on-error))))
 
 ;;; Read and compile the source file.
 (defun sub-sub-compile-file (info)
@@ -963,7 +1001,25 @@ necessary, since type inference may take arbitrarily long to converge.")
     (with-source-paths
       (find-source-paths form current-index)
       (process-toplevel-form
-       form `(original-source-start 0 ,current-index) nil))))
+       form `(original-source-start 0 ,current-index) nil)))
+  ;; It's easy to get into a situation where cold-init crashes and the only
+  ;; backtrace you get from ldb is TOP-LEVEL-FORM, which means you're anywhere
+  ;; within the 23000 or so blobs of code deferred until cold-init.
+  ;; Seeing each file finish narrows things down without the noise of :sb-show,
+  ;; but this hack messes up form positions, so it's not on unless asked for.
+  #+nil ; change to #+sb-xc-host if desired
+  (let ((file-info (get-toplevelish-file-info info)))
+    (declare (ignorable file-info))
+    (let* ((forms (file-info-forms file-info))
+           (form
+            `(write-string
+              ,(format nil "Completed TLFs: ~A~%" (file-info-name file-info))))
+           (index
+            (+ (fill-pointer forms) (file-info-source-root file-info))))
+      (with-source-paths
+        (find-source-paths form index)
+        (process-toplevel-form
+         form `(original-source-start 0 ,index) nil)))))
 
 ;;; Return the INDEX'th source form read from INFO and the position
 ;;; where it was read.
@@ -979,24 +1035,24 @@ necessary, since type inference may take arbitrarily long to converge.")
 ;;; actually compile something. If *BLOCK-COMPILE* is T, then we still
 ;;; convert the form, but delay compilation, pushing the result on
 ;;; *TOPLEVEL-LAMBDAS* instead.
-(defun convert-and-maybe-compile (form path)
+(defun convert-and-maybe-compile (form path &optional (expand t))
   (declare (list path))
   (let ((*top-level-form-noted* (note-top-level-form form t)))
     ;; Don't bother to compile simple objects that just sit there.
     (when (and form (or (symbolp form) (consp form)))
-      (if (fopcompilable-p form)
-         (let ((*fopcompile-label-counter* 0))
-           (fopcompile form path nil))
-         (with-ir1-namespace
-           (let ((*lexenv* (make-lexenv
-                            :policy *policy*
-                            :handled-conditions *handled-conditions*
-                            :disabled-package-locks *disabled-package-locks*))
-                 (tll (ir1-toplevel form path nil)))
-             (if (eq *block-compile* t)
-                 (push tll *toplevel-lambdas*)
-                 (compile-toplevel (list tll) nil))
-             nil))))))
+      (if (fopcompilable-p form expand)
+          (let ((*fopcompile-label-counter* 0))
+            (fopcompile form path nil expand))
+          (with-ir1-namespace
+            (let ((*lexenv* (make-lexenv
+                             :policy *policy*
+                             :handled-conditions *handled-conditions*
+                             :disabled-package-locks *disabled-package-locks*))
+                  (tll (ir1-toplevel form path nil)))
+              (if (eq *block-compile* t)
+                  (push tll *toplevel-lambdas*)
+                  (compile-toplevel (list tll) nil))
+              nil))))))
 
 ;;; Macroexpand FORM in the current environment with an error handler.
 ;;; We only expand one level, so that we retain all the intervening
@@ -1238,18 +1294,6 @@ necessary, since type inference may take arbitrarily long to converge.")
 
               (mapc #'clear-ir1-info components-from-dfo))))))))
 
-(defun process-toplevel-cold-fset (name lambda-expression path)
-  (unless (producing-fasl-file)
-    (error "can't COLD-FSET except in a fasl file"))
-  (legal-fun-name-or-type-error name)
-  (fasl-dump-cold-fset name
-                       (%compile lambda-expression
-                                 *compile-object*
-                                 :name name
-                                 :path path)
-                       *compile-object*)
-  (values))
-
 (defun note-top-level-form (form &optional finalp)
   (when *compile-print*
     (cond ((not *top-level-form-noted*)
@@ -1278,10 +1322,21 @@ necessary, since type inference may take arbitrarily long to converge.")
 ;;; compilation. Normally just evaluate in the appropriate
 ;;; environment, but also compile if outputting a CFASL.
 (defun eval-compile-toplevel (body path)
-  (eval-tlf `(progn ,@body) (source-path-tlf-number path) *lexenv*)
-  (when *compile-toplevel-object*
-    (let ((*compile-object* *compile-toplevel-object*))
-      (convert-and-maybe-compile `(progn ,@body) path))))
+  (flet ((frob ()
+           (eval-tlf `(progn ,@body) (source-path-tlf-number path) *lexenv*)
+           (when *compile-toplevel-object*
+             (let ((*compile-object* *compile-toplevel-object*))
+               (convert-and-maybe-compile `(progn ,@body) path)))))
+    (if (null *macro-policy*)
+        (frob)
+        (let* ((*lexenv*
+                (make-lexenv :policy (process-optimize-decl (macro-policy-decls nil)
+                                                            (lexenv-policy *lexenv*))
+                             :default *lexenv*))
+               ;; In case a null lexenv is created, it needs to get the newly
+               ;; effective global policy, not the policy currently in *POLICY*.
+               (*policy* (lexenv-policy *lexenv*)))
+          (frob)))))
 
 ;;; Process a top level FORM with the specified source PATH.
 ;;;  * If this is a magic top level form, then do stuff.
@@ -1363,7 +1418,7 @@ necessary, since type inference may take arbitrarily long to converge.")
                    (cond ((eq expanded form)
                           (when compile-time-too
                             (eval-compile-toplevel (list form) path))
-                          (convert-and-maybe-compile form path))
+                          (convert-and-maybe-compile form path nil))
                          (t
                           (process-toplevel-form expanded
                                                  path
@@ -1384,16 +1439,6 @@ necessary, since type inference may take arbitrarily long to converge.")
                                        (car form)
                                        form))))
               (case (car form)
-                ;; In the cross-compiler, top level COLD-FSET arranges
-                ;; for static linking at cold init time.
-                #+sb-xc-host
-                ((cold-fset)
-                 (aver (not compile-time-too))
-                 (destructuring-bind (cold-fset fun-name lambda-expression) form
-                   (declare (ignore cold-fset))
-                   (process-toplevel-cold-fset fun-name
-                                               lambda-expression
-                                               path)))
                 ((eval-when macrolet symbol-macrolet);things w/ 1 arg before body
                  (need-at-least-one-arg form)
                  (destructuring-bind (special-operator magic &rest body) form
@@ -1450,6 +1495,18 @@ necessary, since type inference may take arbitrarily long to converge.")
 ;;; Compile FORM and arrange for it to be called at load-time. Return
 ;;; the dumper handle and our best guess at the type of the object.
 (defun compile-load-time-value (form)
+  ;; Special case for the cross-compiler. While the normal ltv stuff is fine
+  ;; for the most part, it is inadequate for SETUP-PRINTER-STATE.
+  ;; In cold-init we want the printer to work before regular ltv forms are run,
+  ;; so this is fop-based magic, slightly dangerous in that it can produce a
+  ;; use of #'F before the referenced function has been defined.
+  ;; Just be careful not to do that.
+  #+sb-xc-host
+  (when (typep form '(cons (eql function) (cons symbol null)))
+    (fopcompile form nil t)
+    (return-from compile-load-time-value
+      (values (sb!fasl::dump-pop *compile-object*)
+              (specifier-type 'function))))
   (let ((lambda (compile-load-time-stuff form t)))
     (values
      (fasl-dump-load-time-value-lambda lambda *compile-object*)
@@ -1631,6 +1688,7 @@ necessary, since type inference may take arbitrarily long to converge.")
         (sb!xc:*compile-file-pathname* nil) ; really bound in
         (sb!xc:*compile-file-truename* nil) ; SUB-SUB-COMPILE-FILE
         (*policy* *policy*)
+        (*macro-policy* *macro-policy*)
         (*code-coverage-records* (make-hash-table :test 'equal))
         (*code-coverage-blocks* (make-hash-table :test 'equal))
         (*handled-conditions* *handled-conditions*)
@@ -1813,7 +1871,9 @@ SPEED and COMPILATION-SPEED optimization values, and the
          (warnings-p nil)
          (failure-p t) ; T in case error keeps this from being set later
          (input-pathname (verify-source-file input-file))
-         (source-info (make-file-source-info input-pathname external-format))
+         (source-info
+          (make-file-source-info input-pathname external-format
+                                 #-sb-xc-host t)) ; can't track, no SBCL streams
          (*compiler-trace-output* nil)) ; might be modified below
 
     (unwind-protect
@@ -1992,6 +2052,13 @@ SPEED and COMPILATION-SPEED optimization values, and the
                 (sb!xc:make-load-form constant (make-null-lexenv))
               (error (condition)
                 (compiler-error condition))))
+      #-sb-xc-host
+      (when (and (not namep)
+                 (listp creation-form) ; skip if already a magic keyword
+                 (typep constant 'structure-object)
+                 (sb!kernel::canonical-slot-saving-forms-p
+                  constant creation-form init-form))
+        (setq creation-form :sb-just-dump-it-normally))
       (case creation-form
         (:sb-just-dump-it-normally
          (fasl-validate-structure constant *compile-object*)
@@ -2011,8 +2078,8 @@ SPEED and COMPILATION-SPEED optimization values, and the
                  (catch constant
                    (fasl-note-handle-for-constant
                     constant
-                    (compile-load-time-value
-                     creation-form)
+                    (or (fopcompile-allocate-instance creation-form)
+                        (compile-load-time-value creation-form))
                     *compile-object*)
                    nil)
                (compiler-error "circular references in creation form for ~S"
@@ -2024,11 +2091,13 @@ SPEED and COMPILATION-SPEED optimization values, and the
                        (loop for (name form) on (cdr info) by #'cddr
                          collect name into names
                          collect form into forms
-                         finally (compile-make-load-form-init-forms forms))
+                         finally (or (fopcompile-constant-init-forms forms)
+                                     (compile-make-load-form-init-forms forms)))
                        nil)))
                (when circular-ref
                  (setf (cdr circular-ref)
-                       (append (cdr circular-ref) (cdr info))))))))))))
+                       (append (cdr circular-ref) (cdr info)))))))
+         nil)))))
 
 
 ;;;; Host compile time definitions
