@@ -34,9 +34,11 @@
 ;;; Return a new POLICY containing the policy information represented
 ;;; by the optimize declaration SPEC. Any parameters not specified are
 ;;; defaulted from the POLICY argument.
-(declaim (ftype (function (list policy) policy) process-optimize-decl))
+(declaim (ftype (function (list policy) (values policy list))
+                process-optimize-decl))
 (defun process-optimize-decl (spec policy)
-  (let ((result nil))
+  (let ((result nil)
+        (specified-qualities))
     ;; Add new entries from SPEC.
     (dolist (q-and-v-or-just-q (cdr spec))
       (multiple-value-bind (quality raw-value)
@@ -59,14 +61,18 @@
                (when (eql quality 'inhibit-warnings)
                  (compiler-style-warn "~S is deprecated: use ~S instead"
                                       quality 'muffle-conditions))
-               (push (cons quality raw-value)
-                     result)))))
+               (let* ((pair (cons quality raw-value))
+                      (found (member quality result :key #'car)))
+                 (push pair specified-qualities)
+                 (if found
+                     (rplaca found pair)
+                     (push pair result)))))))
     ;; Add any nonredundant entries from old POLICY.
     (dolist (old-entry policy)
       (unless (assq (car old-entry) result)
         (push old-entry result)))
     ;; Voila.
-    (sort-policy result)))
+    (values (sort-policy result) specified-qualities)))
 
 (declaim (ftype (function (list list) list)
                 process-handle-conditions-decl))
@@ -223,11 +229,15 @@
 (defun process-inline-declaration (name kind)
   ;; since implicitly it is a function, also scrubs *FREE-FUNS*
   (proclaim-as-fun-name name)
-  (setf (info :function :inlinep name)
-        (ecase kind
+  ;; Check for problems before touching globaldb,
+  ;; so that the report function can see the old value.
+  (let ((newval
+         (ecase kind
           (inline :inline)
           (notinline :notinline)
           (maybe-inline :maybe-inline))))
+    (warn-if-inline-failed/proclaim name newval)
+    (setf (info :function :inlinep name) newval)))
 
 (defun process-declaration-declaration (name form)
   (unless (symbolp name)
@@ -263,13 +273,17 @@
           (t
            decl-spec))))
 
-(defun sb!xc:proclaim (raw-form)
-  #+sb-xc (/show0 "entering PROCLAIM, RAW-FORM=..")
-  #+sb-xc (/hexstr raw-form)
+(defun %proclaim (raw-form location)
   (destructuring-bind (&whole form &optional kind &rest args)
       (canonized-decl-spec raw-form)
-    (labels ((map-names (names function &rest extra-args)
+    (labels ((store-location (name)
+               (if location
+                   (setf (getf (info :source-location :declaration name) kind)
+                         location)
+                   (remf (info :source-location :declaration name) kind)))
+             (map-names (names function &rest extra-args)
                (mapc (lambda (name)
+                       (store-location name)
                        (apply function name extra-args))
                      names))
              (map-args (function &rest extra-args)
@@ -293,7 +307,11 @@
         (freeze-type
          (map-args #'process-freeze-type-declaration))
         (optimize
-         (setq *policy* (process-optimize-decl form *policy*)))
+         (multiple-value-bind (new-policy specified-qualities)
+             (process-optimize-decl form *policy*)
+           (setq *policy* new-policy)
+           (advise-if-repeated-optimize-qualities
+            new-policy specified-qualities)))
         (muffle-conditions
          (setq *handled-conditions*
                (process-muffle-conditions-decl form *handled-conditions*)))
@@ -309,6 +327,24 @@
          (map-args #'process-declaration-declaration form))
         (t
          (unless (info :declaration :recognized kind)
-           (compiler-warn "unrecognized declaration ~S" raw-form))))))
+           (compiler-warn "unrecognized declaration ~S" raw-form)))))))
+
+(defun sb!xc:proclaim (raw-form)
+  #+sb-xc (/show0 "entering PROCLAIM, RAW-FORM=..")
+  #+sb-xc (/hexstr raw-form)
+  (%proclaim raw-form nil)
   #+sb-xc (/show0 "returning from PROCLAIM")
   (values))
+
+(defun advise-if-repeated-optimize-qualities (new-policy specified-qualities)
+  (let (dups)
+    (dolist (quality-and-value specified-qualities)
+      (let* ((quality (car quality-and-value))
+             (current (assq quality new-policy)))
+        (when (and (not (eql (cdr quality-and-value) (cdr current)))
+                   (not (assq quality dups)))
+          (push `(,quality ,(cdr current)) dups))))
+    (when dups
+      (compiler-style-warn
+       "Repeated OPTIMIZE qualit~@P. Using ~{~S~^ and ~}"
+       (length dups) dups))))
