@@ -52,6 +52,7 @@
                `(defmethod ,name ,args
                  (declare (ignore initargs))
                  (error 'metaobject-initialization-violation
+                  ;; FIXME: I'm pretty sure this wants to be "~~@<~A~~@:>"
                   :format-control ,(format nil "~@<~A~@:>" control)
                   :format-arguments (list ',name)
                   :references (list '(:amop :initialization method))))))
@@ -676,7 +677,7 @@
    generic-function
    (types-from-args generic-function classes 'class-eq)))
 
-(defun proclaim-incompatible-superclasses (classes)
+(defun !proclaim-incompatible-superclasses (classes)
   (setq classes (mapcar (lambda (class)
                           (if (symbolp class)
                               (find-class class)
@@ -696,7 +697,7 @@
           (return-from superclasses-compatible-p nil))))))
 
 (mapc
- #'proclaim-incompatible-superclasses
+ #'!proclaim-incompatible-superclasses
  '(;; superclass class
    (system-class std-class structure-class) ; direct subclasses of pcl-class
    (standard-class funcallable-standard-class)
@@ -751,7 +752,6 @@
     (declare (ignore applyp metatypes nkeys))
     (let ((types-rev nil))
       (dotimes-fixnum (i nreq)
-        i
         (unless arguments
           (error-need-at-least-n-args (generic-function-name generic-function)
                                       nreq))
@@ -1424,8 +1424,10 @@
                         (make-dfun-lambda-list nargs applyp)
                         (make-fast-method-call-lambda-list nargs applyp))))
       (multiple-value-bind (cfunction constants)
-          (get-fun1 `(named-lambda (gf-dispatch ,name)
-                      ,arglist
+          ;; We don't want NAMED-LAMBDA for any expressions handed to FNGEN,
+          ;; because name mismatches will render the hashing ineffective.
+          (get-fun1 `(lambda ,arglist
+                      (declare (optimize (sb-c::store-closure-debug-pointer 3)))
                       ,@(unless function-p
                           `((declare (ignore .pv. .next-method-call.))))
                       (locally (declare #.*optimize-speed*)
@@ -1452,7 +1454,7 @@
                 (setf alist-tail new)))
             (let ((function (apply cfunction (mapcar #'cdr (cdr alist)))))
               (if function-p
-                  function
+                  (set-fun-name function `(gf-dispatch ,name))
                   (make-fast-method-call
                    :function (set-fun-name function `(sdfun-method ,name))
                    :arg-info fmc-arg-info))))))))))
