@@ -17,7 +17,6 @@
 #ifndef LISP_FEATURE_WIN32
 #include <sched.h>
 #endif
-#include "runtime.h"
 #include <stddef.h>
 #include <errno.h>
 #include <sys/types.h>
@@ -49,15 +48,27 @@
 #include "interrupt.h"
 #include "lispregs.h"
 
-#if defined(LISP_FEATURE_WIN32) && defined(LISP_FEATURE_SB_THREAD)
+#ifdef LISP_FEATURE_SB_THREAD
+
+#ifdef LISP_FEATURE_OPENBSD
+#include <pthread_np.h>
+#endif
+
+#ifdef LISP_FEATURE_SUNOS
+#include <thread.h>
+#endif
+
+#ifdef LISP_FEATURE_WIN32
 # define IMMEDIATE_POST_MORTEM
 #endif
 
-#if defined(LISP_FEATURE_DARWIN) && defined(LISP_FEATURE_SB_THREAD)
+#ifdef LISP_FEATURE_DARWIN
 #define DELAY_THREAD_POST_MORTEM 5
 #define LOCK_CREATE_THREAD
 #endif
 
+#endif
+
 #if defined(LISP_FEATURE_FREEBSD) || defined(LISP_FEATURE_DRAGONFLY)
 #define LOCK_CREATE_THREAD
 #endif
@@ -495,7 +506,6 @@ new_thread_trampoline(struct thread *th)
     return result;
 }
 
-# ifdef LISP_FEATURE_SB_SAFEPOINT
 static struct thread *create_thread_struct(lispobj);
 
 void
@@ -515,12 +525,32 @@ attach_os_thread(init_thread_data *scribble)
 #ifndef LISP_FEATURE_WIN32
     /* On windows, arch_os_thread_init will take care of finding the
      * stack. */
-    pthread_attr_t attr;
-    int pthread_getattr_np(pthread_t, pthread_attr_t *);
-    pthread_getattr_np(os, &attr);
     void *stack_addr;
     size_t stack_size;
+#ifdef LISP_FEATURE_OPENBSD
+    stack_t stack;
+    pthread_stackseg_np(os, &stack);
+    stack_size = stack.ss_size;
+    stack_addr = (void*)((size_t)stack.ss_sp - stack_size);
+#elif defined LISP_FEATURE_SUNOS
+  stack_t stack;
+  thr_stksegment(&stack);
+  stack_size = stack.ss_size;
+  stack_addr = (void*)((size_t)stack.ss_sp - stack_size);
+#elif defined(LISP_FEATURE_DARWIN)
+    stack_addr = pthread_get_stackaddr_np(os);
+    stack_size = pthread_get_stacksize_np(os);
+#else
+    pthread_attr_t attr;
+#ifdef LISP_FEATURE_FREEBSD
+    pthread_attr_get_np(os, &attr);
+#else
+    int pthread_getattr_np(pthread_t, pthread_attr_t *);
+    pthread_getattr_np(os, &attr);
+#endif
     pthread_attr_getstack(&attr, &stack_addr, &stack_size);
+#endif
+
     th->control_stack_start = stack_addr;
     th->control_stack_end = (void *) (((uintptr_t) stack_addr) + stack_size);
 #endif
@@ -554,8 +584,45 @@ detach_os_thread(init_thread_data *scribble)
     pthread_setspecific(lisp_thread, (void *)0);
     thread_sigmask(SIG_SETMASK, &scribble->oldset, 0);
 }
-# endif /* safepoint */
 
+void
+callback_wrapper_trampoline(
+#if !(defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64))
+    /* On the x86oid backends, the assembly wrapper happens to not pass
+     * in ENTER_ALIEN_CALLBACK explicitly for safepoints.  However, the
+     * platforms with precise GC are tricky enough already, and I want
+     * to minimize the read-time conditionals.  For those platforms, I'm
+     * only replacing funcall3 with callback_wrapper_trampoline while
+     * keeping the arguments unchanged. --DFL */
+    lispobj __attribute__((__unused__)) fun,
+#endif
+    lispobj arg0, lispobj arg1, lispobj arg2)
+{
+#if defined(LISP_FEATURE_WIN32)
+    pthread_np_notice_thread();
+#endif
+    struct thread* th = arch_os_get_current_thread();
+    if (!th) {                  /* callback invoked in non-lisp thread */
+        init_thread_data scribble;
+        attach_os_thread(&scribble);
+        funcall3(StaticSymbolFunction(ENTER_FOREIGN_CALLBACK), arg0,arg1,arg2);
+        detach_os_thread(&scribble);
+        return;
+    }
+
+#ifdef LISP_FEATURE_WIN32
+    /* arg2 is the pointer to a return value, which sits on the stack */
+    th->carried_base_pointer = (os_context_register_t) *(((void**)arg2)-1);
+#endif
+
+#ifdef LISP_FEATURE_SB_SAFEPOINT
+    WITH_GC_AT_SAFEPOINTS_ONLY()
+#endif
+    {
+       funcall3(SymbolValue(ENTER_ALIEN_CALLBACK, 0), arg0, arg1, arg2);
+
+    }
+}
 #endif /* LISP_FEATURE_SB_THREAD */
 
 static void
