@@ -311,3 +311,34 @@
     (type-error (condition)
       (assert (search (string 'no-such-condition)
                       (princ-to-string condition))))))
+
+;; Using an undefined condition type in a HANDLER-BIND clause should
+;; signal an ERROR at runtime. Bug 1378939 was about landing in LDB
+;; because of infinite recursion in SIGNAL instead.
+;;
+;; We suppress the compile-time WARNING to avoid noise when running
+;; tests.
+(locally (declare (muffle-conditions warning))
+  (with-test (:name (handler-bind :undefined-condition-type
+                     :bug-1378939))
+    (assert-error
+     (handler-bind ((no-such-condition-class #'print))
+       (error "does not matter")))))
+
+;; Using an undefined condition type in a HANDLER-BIND clause should
+;; signal a [STYLE-]WARNING at compile time.
+(with-test (:name (handler-bind :undefined-condition-type
+                   :compile-time-warning))
+  (handler-bind ((warning #'muffle-warning))
+    (assert-signal
+     (compile nil '(lambda () (handler-bind
+                                  ((no-such-condition-class #'print)))))
+     warning)))
+
+;; Empty bindings in HANDLER-BIND pushed an empty cluster onto
+;; *HANDLER-CLUSTERS* which was not expected by SIGNAL (and wasteful).
+(with-test (:name (handler-bind :empty-bindings :bug-1388707))
+  (multiple-value-bind (value condition)
+      (ignore-errors (handler-bind () (error "Foo")))
+    (assert (null value))
+    (assert (typep condition 'simple-error))))
