@@ -171,10 +171,10 @@
          (declare (optimize (speed 2) (debug 1))) ; tail call elimination
          (#:undefined-function 42))
        (not-optimized ()
-         (declare (optimize (speed 1) (debug 2))) ; no tail call elimination
+         (declare (optimize (speed 1) (debug 3))) ; no tail call elimination
          (#:undefined-function 42))
        (test (fun)
-         (declare (optimize (speed 1) (debug 2))) ; no tail call elimination
+         (declare (optimize (speed 1) (debug 3))) ; no tail call elimination
          (funcall fun)))
 
   (with-test (:name (:undefined-function :bug-346)
@@ -232,10 +232,10 @@
          (declare (optimize (speed 2) (debug 1))) ; tail call elimination
          (/ 42 0))
        (not-optimized ()
-         (declare (optimize (speed 1) (debug 2))) ; no tail call elimination
+         (declare (optimize (speed 1) (debug 3))) ; no tail call elimination
          (/ 42 0))
        (test (fun)
-         (declare (optimize (speed 1) (debug 2))) ; no tail call elimination
+         (declare (optimize (speed 1) (debug 3))) ; no tail call elimination
          (funcall fun)))
   (with-test (:name (:divide-by-zero :bug-346)
               :fails-on :alpha)  ; bug 346
@@ -289,7 +289,7 @@
                                '((,two-arg 42 t)
                                  #+(or x86 x86-64)
                                  ,@(and predicate
-                                    '(("no debug information for frame")))
+                                    `((,(find-symbol (format nil "GENERIC-~A" fun) "SB-VM"))))
                                  ((flet ,test-name :in ,*p*) 42 t))))))))
              (test-predicates (&rest functions)
                `(progn ,@(mapcar (lambda (function)
@@ -827,6 +827,19 @@
       (unless (zerop (length problems))
         (error problems)))))
 
+;; The test named :GF-dispatch-backtrace depends on the fact that renaming
+;; a closure works, and that the debugger can extract a closure name.
+;; First things first: verify that a closure can be named.
+(defun make-adder (x)
+  (sb-impl::set-closure-name (lambda (y) (+ x y)) `(adder ,x)))
+(with-test (:name :closure-renaming-really-works)
+  (let ((f1 (make-adder 5))
+        (expect "#<CLOSURE (ADDER 5)"))
+    (assert (= (mismatch (write-to-string (make-adder 5)) expect)
+               (length expect)))
+    (assert (and (eq (sb-impl::set-closure-name f1 "ADD5") f1)
+                 (string= (sb-impl::%fun-name f1) "ADD5")))))
+
 (defgeneric gf-dispatch-test/gf (x y)
   (:method (x y)
     (+ x y)))
@@ -873,4 +886,210 @@
          (position (+ (search call printed) (length call))))
     (assert (eql position (search "((#))" printed :start2 position)))))
 
-(write-line "/debug.impure.lisp done")
+
+(defvar *x* nil)
+(defun foo (a) a)
+
+(with-test (:name :trace-debug-arg)
+  (trace foo :print-after (setf *x* (sb-debug:arg 0)))
+  (foo 1)
+  (assert (eql *x* 1))
+
+  (trace foo :print (setf *x* (sb-debug:arg 0)))
+  (foo 2)
+  (assert (eql *x* 2))
+
+  (trace foo :condition (eql (setf *x* (sb-debug:arg 0)) 0))
+  (foo 3)
+  (assert (eql *x* 3))
+
+  (trace foo :condition-after (setf *x* (sb-debug:arg 0)))
+  (foo 4)
+  (assert (eql *x* 4))
+
+  (trace foo :break (and (setf *x* (sb-debug:arg 0)) nil))
+  (foo 5)
+  (assert (eql *x* 5))
+
+  (trace foo :break-all (and (setf *x* (sb-debug:arg 0)) nil))
+  (foo 6)
+  (assert (eql *x* 6))
+  (trace foo :break-after (and (setf *x* (sb-debug:arg 0)) nil))
+  (foo 7))
+
+(defun frobbleize (arg) (sb-debug:print-backtrace) 'win)
+(defmethod low-debug-method ((self t))
+  (declare (optimize (debug 0)))
+  (frobbleize 'me) ; make this not a tail call, so it remains on stack
+  5)
+(with-test (:name :clean-fast-method-frame-lossage)
+  (low-debug-method 42)) ; no need to assert. it either crashes or doesn't
+
+(defun return-65535 ()
+  65535)
+
+(with-test (:name :indirect-closure-values)
+  (let ((count 0))
+    (block nil
+      (handler-bind ((error (lambda (c)
+                              (declare (ignore c))
+                              (sb-debug::map-backtrace
+                               (lambda (frame)
+                                 (let ((sb-debug::*current-frame* frame)
+                                       (name (sb-debug::frame-call frame)))
+                                   (when (or (eq name 'test)
+                                             (and (consp name)
+                                                  (or (eql (search '(labels f1) name) 0)
+                                                      (eql (search '(labels f2) name) 0))))
+                                     (incf count)
+                                     (assert (eql (var 'a) 2))))))
+                              (return))))
+        (funcall
+         (compile nil
+                  `(sb-int:named-lambda test ()
+                     (declare (optimize debug))
+                     (let ((a 1))
+                       (labels
+                           ((f1 ()
+                              (incf a)
+                              (signal 'error))
+                            (f2 ()
+                              (f1)))
+                         (f2))))))))
+    (assert (= count 3))))
+
+(with-test (:name :indirect-closure-values.2)
+  (let ((count 0))
+    (block nil
+      (handler-bind ((error (lambda (c)
+                              (declare (ignore c))
+                              (sb-debug::map-backtrace
+                               (lambda (frame)
+                                 (let ((sb-debug::*current-frame* frame)
+                                       (name (sb-debug::frame-call frame)))
+                                   (when (or (eq name 'test)
+                                             (and (consp name)
+                                                  (or (eql (search '(labels f1) name) 0)
+                                                      (eql (search '(labels f2) name) 0))))
+                                     (incf count)
+                                     (assert (eql (var 'a) 65535))))))
+                              (return))))
+        (funcall
+         (compile nil
+                  `(sb-int:named-lambda test ()
+                     (declare (optimize debug))
+                     (let ((a (return-65535)))
+                       (declare ((unsigned-byte 16) a))
+                       (labels
+                           ((f1 ()
+                              (incf a)
+                              (signal 'error))
+                            (f2 ()
+                              (f1)))
+                         (f2))))))))
+    (assert (= count 3))))
+
+(with-test (:name :non-tail-self-call-bad-variables)
+  (let ((count 0))
+    (block nil
+      (handler-bind ((error (lambda (c)
+                              (declare (ignore c))
+                              (sb-debug::map-backtrace
+                               (lambda (frame)
+                                 (let ((sb-debug::*current-frame* frame))
+                                   (multiple-value-bind (name args)
+                                       (sb-debug::frame-call frame)
+                                     (when (eq name 'test)
+                                       (assert (or (null args)
+                                                   (equal args '(nil))))
+                                       (incf count))))))
+                              (return))))
+        (funcall
+         (compile nil `(sb-int:named-lambda test (&optional x)
+                         (declare (optimize sb-c::recognize-self-calls))
+                         (signal 'error :format-control "~a" :format-arguments (list x))
+                         (test 1)
+                         1)))))
+    (assert (= count 1))))
+
+(with-test (:name :local-tail-call)
+  (assert (verify-backtrace
+           (lambda () (funcall (compile nil `(sb-int:named-lambda test ()
+                                               (signal 'error)
+                                               (flet ((tail ()))
+                                                 (declare (notinline tail))
+                                                 (tail))))))
+           '((test)))))
+
+(with-test (:name :local-tail-call-variables)
+  (let ((count 0))
+    (block nil
+      (handler-bind ((error (lambda (c)
+                              (declare (ignore c))
+                              (sb-debug::map-backtrace
+                               (lambda (frame)
+                                 (let ((sb-debug::*current-frame* frame))
+                                   (multiple-value-bind (name args)
+                                       (sb-debug::frame-call frame)
+                                     (when (eq name 'test)
+                                       (assert (equal args '(error)))
+                                       (incf count))))))
+                              (return))))
+        (funcall
+         (compile nil `(sb-int:named-lambda test (x)
+                         (signal x)
+                         ;; If :local-tail-call fails, this will fail
+                         ;; too, because there's no jump between
+                         ;; SIGNAL and the call to TAIL and it will
+                         ;; show (flet tail) in the backtrace.
+                         (flet ((tail ()))
+                           (declare (notinline tail))
+                           (tail))))
+         'error)))
+    (assert (= count 1))))
+
+(with-test (:name :variables-surrounding-inlined-code)
+  (let ((count 0))
+    (block nil
+      (handler-bind ((error (lambda (c)
+                              (declare (ignore c))
+                              (sb-debug::map-backtrace
+                               (lambda (frame)
+                                 (let ((sb-debug::*current-frame* frame))
+                                   (multiple-value-bind (name)
+                                       (sb-debug::frame-call frame)
+                                     (when (eq name 'test)
+                                       (assert (equal (sb-debug:var 'l) '(1 2 3)))
+                                       (incf count))))))
+                              (return))))
+        (funcall
+         (compile nil `(sb-int:named-lambda test (a i)
+                         (declare (optimize (debug 3)))
+                         (let ((l (list 1 2 3)))
+                           (aref a i)
+                           l)))
+         #(1) 2)))
+    (assert (= count 1))))
+
+(with-test (:name :variables-surrounding-inlined-code.2)
+  (let ((count 0))
+    (block nil
+      (handler-bind ((error (lambda (c)
+                              (declare (ignore c))
+                              (sb-debug::map-backtrace
+                               (lambda (frame)
+                                 (let ((sb-debug::*current-frame* frame))
+                                   (multiple-value-bind (name)
+                                       (sb-debug::frame-call frame)
+                                     (when (eq name 'test)
+                                       (assert (equal (sb-debug:var 'l) '(1 2 3)))
+                                       (incf count))))))
+                              (return))))
+        (funcall
+         (compile nil `(sb-int:named-lambda test (c)
+                         (declare (optimize (debug 3)))
+                         (let ((l (list 1 2 3)))
+                           (map 'list #'signal c)
+                           l)))
+         '(error))))
+    (assert (= count 1))))
