@@ -75,10 +75,10 @@
 
 (test-util:with-test (:name :set-info-value-type-check)
   (loop for type-info across *info-types*
-        when (and type-info (not (eq (type-info-type-spec type-info) 't)))
+        when (and type-info (not (eq (meta-info-type-spec type-info) 't)))
         do
-        (let ((key1 (type-info-class type-info))
-              (key2 (type-info-name type-info))
+        (let ((key1 (meta-info-category type-info))
+              (key2 (meta-info-kind type-info))
               (sillyval (make-string-output-stream))) ; nothing should be this
           ;; check the type-checker function
           (let ((f (compile nil
@@ -96,7 +96,7 @@
                            'type-error)))))
   ;; but if I *really* want, a bad value can be installed
   (set-info-value (gensym)
-                  (type-info-number (type-info-or-lose :variable :kind))
+                  (meta-info-number (meta-info-or-lose :variable :kind))
                   :this-is-no-good))
 
 (test-util:with-test (:name :unrecognize-recognized-declaration)
@@ -179,16 +179,16 @@
   (mapcan (lambda (x) (mapcar (lambda (y) (cons x y)) b))
           a))
 
-;; The real GET-INFO-VALUE AVERs that TYPE-NUMBER is legal. This one doesn't.
-(defun cheating-get-info-value (sym aux-key type-number)
+;; The real GET-INFO-VALUE AVERs that INFO-NUMBER is legal. This one doesn't.
+(defun cheating-get-info-value (sym aux-key info-number)
   (let* ((vector (symbol-info-vector sym))
-         (index (packed-info-value-index vector aux-key type-number)))
+         (index (packed-info-value-index vector aux-key info-number)))
     (if index
         (values (svref vector index) t)
         (values nil nil))))
 
 ;; Info vectors may be concurrently updated. If more than one thread writes
-;; the same name/type-number, it's random which thread prevails, but for
+;; the same name/info-number, it's random which thread prevails, but for
 ;; non-colliding updates, none should be lost.
 ;; This is not a "reasonable" use of packed info vectors.
 ;; It's just a check of the response of the algorithm to heavy pounding.
@@ -365,7 +365,7 @@
 (defun classoid-cell-test-get-lotsa-symbols ()
   (remove-if-not
    #'symbolp
-   (package-hashtable-table
+   (package-hashtable-cells
     (package-internal-symbols (find-package "SB-C")))))
 
 ;; Make every symbol in the test set have a classoid-cell
@@ -465,7 +465,7 @@
         (n-created 0)
         (highest-type-num
          (position-if #'identity sb-c::*info-types*
-                      :end sb-c::+fdefn-type-num+ :from-end t)))
+                      :end sb-int:+fdefn-info-num+ :from-end t)))
     (loop for name across names
           for i from 0
           do (setf (aref fdefn-result i)
